import rospy
import gym
from gym.utils import seeding
from .gazebo_connection import GazeboConnection
from .controllers_connection import ControllersConnection
#https://bitbucket.org/theconstructcore/theconstruct_msgs/src/master/msg/RLExperimentInfo.msg
from openai_ros.msg import RLExperimentInfo

# https://github.com/openai/gym/blob/master/gym/core.py
class RobotGazeboEnv(gym.Env):

    def __init__(self, robot_name_space, controllers_list, reset_controls, start_init_physics_parameters=True, reset_world_or_sim="SIMULATION"):

        # To reset Simulations
        rospy.logdebug("START init RobotGazeboEnv")
        self.gazebo = GazeboConnection(start_init_physics_parameters,reset_world_or_sim)
        self.controllers_object = ControllersConnection(namespace=robot_name_space, controllers_list=controllers_list)
        self.reset_controls = reset_controls
        self.seed()

        # Set up ROS related variables
        self.episode_num = 0
        self.cumulated_episode_reward = 0
        self.reward_pub = rospy.Publisher('/openai/reward', RLExperimentInfo, queue_size=1)

        # We Unpause the simulation and reset the controllers if needed
        """
        To check any topic we need to have the simulations running, we need to do two things:
        1) Unpause the simulation: without that th stream of data doesnt flow. This is for simulations
        that are pause for whatever the reason
        2) If the simulation was running already for some reason, we need to reset the controlers.
        This has to do with the fact that some plugins with tf, dont understand the reset of the simulation
        and need to be reseted to work properly.
        """
        self.gazebo.unpauseSim()
        if self.reset_controls:
            self.controllers_object.reset_controllers()

        rospy.logdebug("END init RobotGazeboEnv")

    # Env methods
    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        """
        Function executed each time step.
        Here we get the action execute it in a time step and retrieve the
        observations generated by that action.
        :param action:
        :return: obs, reward, done, info
        """

        """
        Here we should convert the action num to movement action, execute the action in the
        simulation and get the observations result of performing that action.
        """
        rospy.logdebug("START STEP OpenAIROS")

        self.gazebo.unpauseSim()
        self._set_action(action)
        # self.gazebo.pauseSim()
        obs = self._get_obs()
        done = self._is_done(obs)
        info = {}
        reward = self._compute_reward(obs, done)
        self.cumulated_episode_reward += reward

        rospy.logdebug("END STEP OpenAIROS")

        return obs, reward, done, info

    def reset(self):
        rospy.logdebug("Reseting RobotGazeboEnvironment")
        self._reset_sim()
        self._init_env_variables()
        self._update_episode()
        obs = self._get_obs()
        rospy.logdebug("END Reseting RobotGazeboEnvironment")
        return obs

    def close(self):
        """
        Function executed when closing the environment.
        Use it for closing GUIS and other systems that need closing.
        :return:
        """
        rospy.logdebug('Closing RobotGazeboEnvironment')
        rospy.signal_shutdown('Closing RobotGazeboEnvironment')

    def _update_episode(self):
        """
        Publishes the cumulated reward of the episode and
        increases the episode number by one.
        :return:
        """
        rospy.logwarn("PUBLISHING REWARD...")
        self._publish_reward_topic(
                                    self.cumulated_episode_reward,
                                    self.episode_num
                                    )
        rospy.logwarn("PUBLISHING REWARD...DONE="+str(self.cumulated_episode_reward)+",EP="+str(self.episode_num))

        self.episode_num += 1
        self.cumulated_episode_reward = 0


    def _publish_reward_topic(self, reward, episode_number=1):
        """
        This function publishes the given reward in the reward topic for
        easy access from ROS infrastructure.
        :param reward:
        :param episode_number:
        :return:
        """
        reward_msg = RLExperimentInfo()
        reward_msg.episode_number = episode_number
        reward_msg.episode_reward = reward
        self.reward_pub.publish(reward_msg)

    # Extension methods
    # ----------------------------

    def _reset_sim(self):
        """Resets a simulation
        """
        rospy.logdebug("RESET SIM START")
        if self.reset_controls :
            rospy.logdebug("RESET CONTROLLERS")
            self.gazebo.unpauseSim()
            self.controllers_object.reset_controllers()
            self._check_all_systems_ready()
            self._set_init_pose()
            # self.gazebo.pauseSim()
            self.gazebo.resetSim()
            # self.gazebo.unpauseSim()
            self.controllers_object.reset_controllers()
            self._check_all_systems_ready()
            # self.gazebo.pauseSim()

        else:
            rospy.logwarn("DONT RESET CONTROLLERS")
            self.gazebo.unpauseSim()
            self._check_all_systems_ready()
            self._set_init_pose()
            # self.gazebo.pauseSim()
            self.gazebo.resetSim()
            # self.gazebo.unpauseSim()
            self._check_all_systems_ready()
            # self.gazebo.pauseSim()

        rospy.logdebug("RESET SIM END")
        return True

    def _set_init_pose(self):
        """Sets the Robot in its init pose
        """
        raise NotImplementedError()

    def _check_all_systems_ready(self):
        """
        Checks that all the sensors, publishers and other simulation systems are
        operational.
        """
        raise NotImplementedError()

    def _get_obs(self):
        """Returns the observation.
        """
        raise NotImplementedError()

    def _init_env_variables(self):
        """Inits variables needed to be initialised each time we reset at the start
        of an episode.
        """
        raise NotImplementedError()

    def _set_action(self, action):
        """Applies the given action to the simulation.
        """
        raise NotImplementedError()

    def _is_done(self, observations):
        """Indicates whether or not the episode is done ( the robot has fallen for example).
        """
        raise NotImplementedError()

    def _compute_reward(self, observations, done):
        """Calculates the reward to give based on the observations given.
        """
        raise NotImplementedError()

    def _env_setup(self, initial_qpos):
        """Initial configuration of the environment. Can be used to configure initial state
        and extract information from the simulation.
        """
        raise NotImplementedError()




## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.bf
## -- Module  : ml.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-08-20  0.0.0     DA       Creation 
## -- 2021-08-25  1.0.0     DA       Release of first version
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -- 2021-10-06  1.0.1     DA       Extended class Adaptive by new methods _adapt(), get_adapted(),
## --                                _set_adapted(); moved Buffer classes to mlpro.bf.data.py
## -- 2021-10-25  1.0.2     SY       Enhancement of class Adaptive by adding ScientificObject.
## -- 2021-10-26  1.1.0     DA       New class AdaptiveFunction
## -- 2021-10-29  1.1.1     DA       New method Adaptive.set_random_seed()
## -- 2021-11-15  1.2.0     DA       - Class Adaptive renamed to Model
## --                                - New classes Mode, Scenario, TrainingResults, Training, 
## --                                  HyperParamTuner
## -- 2021-11-30  1.2.1     DA       - Classes Model, AdaptiveFunction: new opt. parameters **p_par
## --                                - Docstrings reformatted to numpy style
## --                                - Class Model: new method get_maturity()
## -- 2021-12-07  1.2.2     DA       - Method Scenario.__init__(): param p_cycle_len removed
## --                                - Method Training.__init__(): par p_scenario replaced by p_scenario_cls
## -- 2021-12-08  1.2.3     DA       Moved class AdaptiveFunction to new subtopic package sl
## -- 2021-12-09  1.2.4     DA       Class Training: introduced dynamic parameters **p_kwargs
## -- 2021-12-10  1.2.5     DA       Class HyperParamTuner: changed interface of method maximize()
## -- 2021-12-12  1.3.0     DA       Classes Scenario, Training, TrainingResults: introduced number
## --                                of adapatations
## -- 2021-12-21  1.3.1     DA       - Minor changes on class Training
## --                                - Added log functionality to class TrainingResults
## -- 2022-01-18  1.3.2     MRD      Small optimize on Scenario instantiation in Training class
## --                                Put the self._cycle_limit directly on the parameter argument
## -- 2022-01-27  1.3.3     SY       Class Training: enhanced training with hyperparameter tuning
## -- 2022-01-28  1.3.4     SY       Class HyperParamTuner: add save(), save_line(), HPDataStoring
## -- 2022-02-24  1.3.5     SY       Introduce new class HyperParamDispatcher
## -- 2022-03-02  1.3.6     SY       Refactoring class HyperParamDispatcher
## -- 2022-03-02  1.3.7     DA       Class HyperParamDispatcher:correction of method set_values()
## -- 2022-06-06  1.3.8     MRD      Add additional parameter to Training class, p_env_mode for
## --                                setting up environment mode
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.3.8 (2022-06-06)
This module provides fundamental machine learning templates, functionalities and properties.
"""


import sys
from typing import Dict
from mlpro.bf.various import *
from mlpro.bf.math import *
from mlpro.bf.data import Buffer
from mlpro.bf.plot import *
import random




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class HyperParam (Dimension):
    """
    Hyperparameter definition class. See class Dimension for further descriptions.
    """

## -------------------------------------------------------------------------------------------------
    def register_callback(self, p_cb):
        self._cb = p_cb


## -------------------------------------------------------------------------------------------------
    def callback_on_change(self, p_value):
        try:
            self._cb(p_value)
        except:
            pass





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class HyperParamSpace (ESpace):
    """
    Hyperparameter space, which is just an Euclidian space.
    """

    pass





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class HyperParamTuple (Element):
    """
    Tuple of hyperparameters, which is an element of a hyperparameter space
    """

## -------------------------------------------------------------------------------------------------
    def set_value(self, p_dim_id, p_value):
        super().set_value(p_dim_id, p_value)
        self._set.get_dim(p_dim_id).callback_on_change(p_value)





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class HyperParamDispatcher (HyperParamTuple):
    """
    To dispatch multiple hp tuples into one tuple
    """


## -------------------------------------------------------------------------------------------------
    def __init__(self, p_set: Set) -> None:
        super().__init__(p_set)
        self._hp_dict = {}


## -------------------------------------------------------------------------------------------------
    def add_hp_tuple(self, p_hpt:HyperParamTuple):
        for idx in p_hpt.get_dim_ids():
            self._hp_dict[idx] = p_hpt


## -------------------------------------------------------------------------------------------------
    def get_value(self, p_dim_id):
        return self._hp_dict.get(p_dim_id).get_value(p_dim_id)
   

## -------------------------------------------------------------------------------------------------
    def set_value(self, p_dim_id, p_value):
        self._hp_dict.get(p_dim_id).set_value(p_dim_id, p_value)


## -------------------------------------------------------------------------------------------------
    def get_values(self):
        for idx, dim_id in enumerate(self._set.get_dim_ids()):
            self._values[idx] = self.get_value(dim_id)
        return self._values


## -------------------------------------------------------------------------------------------------
    def set_values(self, p_values):
        for idx, dim_id in enumerate(self._set.get_dim_ids()):
            self.set_value(dim_id, p_values[idx])





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Model (Log, LoadSave, Plottable, ScientificObject):
    """
    Fundamental template class for adaptive ML models. Supports especially
      - Adaptivity
      - Data buffering
      - Hyperparameter management
      - Plotting
      - Scientific referencing on source code level

    Parameters
    ----------
    p_buffer_size : int
        Initial size of internal data buffer. Defaut = 0 (no buffering).
    p_ada : bool
        Boolean switch for adaptivitiy. Default = True.
    p_logging
        Log level (see constants of class Log). Default: Log.C_LOG_ALL
    p_par : Dict
        Futher model specific parameters (to be defined in chhild class).

    """

    C_TYPE          = 'Model'
    C_NAME          = '????'

    C_BUFFER_CLS    = Buffer       

    C_SCIREF_TYPE   = ScientificObject.C_SCIREF_TYPE_NONE     

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_buffer_size=0, p_ada=True, p_logging=Log.C_LOG_ALL, **p_par):  

        Log.__init__(self, p_logging=p_logging)
        self._adapted           = False
        self.switch_adaptivity(p_ada)
        self._hyperparam_space  = HyperParamSpace()
        self._hyperparam_tuple  = None
        self._init_hyperparam(**p_par)

        if p_buffer_size > 0:
            self._buffer = self.C_BUFFER_CLS(p_size=p_buffer_size)
        else:
            self._buffer = None


## -------------------------------------------------------------------------------------------------
    def _init_hyperparam(self, **p_par):
        """
        Implementation specific hyperparameters can be added here. Please follow these steps:
        a) Add each hyperparameter as an object of type HyperParam to the internal hyperparameter
           space object self._hyperparam_space
        b) Create hyperparameter tuple and bind to self._hyperparam_tuple
        c) Set default value for each hyperparameter

        Parameters
        ----------
        p_par : Dict
            Futher model specific parameters, that are passed through constructor.

        """

        pass


## -------------------------------------------------------------------------------------------------
    def get_hyperparam(self) -> HyperParamTuple:
        """
        Returns the internal hyperparameter tuple to get access to single values.
        """

        return self._hyperparam_tuple


## -------------------------------------------------------------------------------------------------
    def switch_adaptivity(self, p_ada:bool):
        """
        Switches adaption functionality on/off.
        
        Parameters
        ----------
        p_ada : bool
            Boolean switch for adaptivity

        """

        self._adaptivity = p_ada
        if self._adaptivity:
            self.log(self.C_LOG_TYPE_I, 'Adaptivity switched on')
        else:
            self.log(self.C_LOG_TYPE_I, 'Adaptivity switched off')


## -------------------------------------------------------------------------------------------------
    def set_random_seed(self, p_seed=None):
        """
        Resets the internal random generator using the given seed.
        """

        random.seed(p_seed)


## -------------------------------------------------------------------------------------------------
    def get_adapted(self) -> bool:
        """
        Returns True, if the model was adapted at least once. False otherwise.
        """

        return self._adapted


## -------------------------------------------------------------------------------------------------
    def _set_adapted(self, p_adapted:bool):
        """
        Sets the adapted flag. Internal use only.
        """

        self._adapted = p_adapted


## -------------------------------------------------------------------------------------------------
    def adapt(self, *p_args) -> bool:
        """
        Adapts the model by calling the custom method _adapt().

        Parameters
        ----------
        p_args
            All parameters that are needed for the adaption. Depends on the specific higher context.

        Returns
        -------
        bool
            True, if something has been adapted. False otherwise.

        """

        if not self._adaptivity: return False
        self.log(self.C_LOG_TYPE_I, 'Adaptation started')
        adapted = self._adapt(*p_args)
        self._set_adapted(adapted)
        return adapted
        

## -------------------------------------------------------------------------------------------------
    def _adapt(self, *p_args) -> bool:
        """
        Custom implementation of the adaptation algorithm. Please describe the type and purpose of 
        all parameters needed by your implementation. This method will be called by public method 
        adapt() if adaptivity is switched on. 

        Parameters
        ----------
        p_args[0]           
            ...
        p_args[1]           
            ...

        Returns
        -------
        bool
            True, if something has been adapted. False otherwise.

        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def clear_buffer(self):
        """
        Clears internal buffer (if buffering is active).
        """

        if self._buffer is not None: self._buffer.clear()


## -------------------------------------------------------------------------------------------------
    def get_maturity(self):
        """
        Computes the maturity of the model.

        Returns
        -------
        float
            Maturity of the model as a scalar value in interval [0,1]
        """

        raise NotImplementedError





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Mode (Log):
    """
    Property class that adds a mode and related methods to a child class.

    Parameters
    ----------
    p_mode
        Operation mode. Valid values are stored in constant C_VALID_MODES.
    p_logging
        Log level (see constants of class Log). Default: Log.C_LOG_ALL

    """

    C_MODE_INITIAL  = -1
    C_MODE_SIM      = 0
    C_MODE_REAL     = 1

    C_VALID_MODES   = [ C_MODE_SIM, C_MODE_REAL ]

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_mode, p_logging=Log.C_LOG_ALL):
        super().__init__(p_logging)
        self._mode = self.C_MODE_INITIAL
        self.set_mode(p_mode)


## -------------------------------------------------------------------------------------------------
    def get_mode(self):
        """
        Returns current mode.
        """

        return self._mode


## -------------------------------------------------------------------------------------------------
    def set_mode(self, p_mode):
        """
        Sets new mode.

        Parameters
        ----------
        p_mode
            Operation mode. Valid values are stored in constant C_VALID_MODES.

        """

        if not p_mode in self.C_VALID_MODES: raise ParamError('Invalid mode')
        if self._mode == p_mode: return
        self._mode = p_mode
        self.log(self.C_LOG_TYPE_I, 'Operation mode set to', self._mode)





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Scenario (Mode, LoadSave, Plottable):
    """
    Template class for a common ML scenario with an adaptive model inside. To be inherited and 
    specialized in higher ML subtopic layers.
    
    The following key features are included:
      - Operation mode
      - Cycle management
      - Timer
      - Latency 
      - Explicit handling of an adaptive ML model inside

    Parameters
    ----------
    p_mode
        Operation mode. See Mode.C_VALID_MODES for valid values. Default = Mode.C_MODE_SIM.
    p_ada : bool
        Boolean switch for adaptivity. Default = True.
    p_cycle_limit : int
        Maximum number of cycles. Default = 0 (no limit).
    p_visualize : bool
        Boolean switch for env/agent visualisation. Default = True.
    p_logging
        Log level (see constants of class Log). Default: Log.C_LOG_ALL.
    
    """

    C_TYPE      = 'Scenario'
    C_NAME      = '????'

## -------------------------------------------------------------------------------------------------
    def __init__(self, 
                 p_mode=Mode.C_MODE_SIM,       
                 p_ada:bool=True,               
                 p_cycle_limit=0,              
                 p_visualize=True,              
                 p_logging=Log.C_LOG_ALL ):    

        # 0 Intro
        self._cycle_max     = sys.maxsize
        self._cycle_id      = 0
        self._visualize     = p_visualize
        self.set_cycle_limit(p_cycle_limit)


        # 1 Setup entire scenario
        self._model = self._setup(p_mode=p_mode, p_ada=p_ada, p_logging=p_logging)
        if self._model is None: 
            raise ImplementationError('Please return your ML model in method self._setup()')

        super().__init__(p_mode, p_logging)


        # 2 Init timer
        if self.get_mode() == Mode.C_MODE_SIM:
            t_mode = Timer.C_MODE_VIRTUAL
        else:
            t_mode = Timer.C_MODE_REAL

        self._cycle_len = self.get_latency()
        self._timer     = Timer(t_mode, self._cycle_len, self._cycle_limit)


## -------------------------------------------------------------------------------------------------
    def switch_logging(self, p_logging):
        super().switch_logging(p_logging)
        self._model.switch_logging(p_logging)


## -------------------------------------------------------------------------------------------------
    def _setup(self, p_mode, p_ada:bool, p_logging) -> Model:
        """
        Custom setup of ML scenario.

        Parameters
        ----------
        p_mode
            Operation mode. See Mode.C_VALID_MODES for valid values. Default = Mode.C_MODE_SIM
        p_ada : bool
            Boolean switch for adaptivity.
        p_logging
            Log level (see constants of class Log). 

        Returns
        -------
        Model
            Adaptive model inside the ML scenario
        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        self._model.init_plot(p_figure=p_figure)


## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        self._model.update_plot()


## -------------------------------------------------------------------------------------------------
    def get_model(self) -> Model:
        """
        Returns the adaptive model object inside the scenario.
        """

        return self._model


## -------------------------------------------------------------------------------------------------
    def set_mode(self, p_mode):
        super().set_mode(p_mode)
        self._set_mode(p_mode)


## -------------------------------------------------------------------------------------------------
    def _set_mode(self, p_mode):
        """
        Redefine this method to switch the scenario between simulation or real operation mode.
        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def get_latency(self) -> timedelta:
        """
        Returns the latency of the scenario. To be implemented in child class.
        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def set_cycle_limit(self, p_limit):
        self._cycle_limit = p_limit


## -------------------------------------------------------------------------------------------------
    def reset(self, p_seed=1):
        """
        Resets the scenario and especially the ML model inside. Internal random generators are seed 
        with the given value. Custom reset actions can be implemented in method _reset().

        Parameters
        ----------
        p_seed : int          
            Seed value for internal random generator

        """

        # 0 Intro
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Scenario reset with seed', str(p_seed))

        # 1 Internal ML model reset
        self._model.set_random_seed(p_seed)
        if self._visualize: self._model.init_plot()

        # 2 Custom reset of further scenario-specific components
        self._reset(p_seed)

        # 3 Timer reset
        self._timer.reset()

        # 4 Cycle counter reset
        self._cycle_id = 0


## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed):
        """
        Custom method to reset further components of the scenario (not the ML model itself) and to 
        set the given random seed value. See method reset() for further details.
        """

        pass


## -------------------------------------------------------------------------------------------------
    def run_cycle(self):
        """
        Runs a single process cycle.

        Returns
        -------
        success : bool
            True on success. False otherwise.
        error : bool
            True on error. False otherwise.
        timeout : bool
            True on timeout. False otherwise.
        cycle_limit : bool
            True, if cycle limit has reached. False otherwise.
        adapted : bool
            True, if ml model adapted something in this cycle. False otherwise.

        """

        # 1 Run a single custom cycle
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Start of cycle', str(self._cycle_id))
        success, error, adapted = self._run_cycle()
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': End of cycle', str(self._cycle_id), '\n')


        # 2 Update visualization
        if self._visualize:
            self.update_plot()


        # 3 Update cycle id and check for optional limit
        if ( self._cycle_limit > 0 ) and ( self._cycle_id >= ( self._cycle_limit -1 ) ): 
            limit = True
        else:
            self._cycle_id = ( self._cycle_id + 1 ) & self._cycle_max
            limit = False


        # 4 Wait for next cycle (real mode only)
        if ( self._timer.finish_lap() == False ) and ( self._cycle_len is not None ):
            self.log(self.C_LOG_TYPE_W, 'Process time', self._timer.get_time(), ': Process timed out !!!')
            timeout = True
        else:
            timeout = False


        # 5 Return result of custom cycle execution
        return success, error, timeout, limit, adapted


## -------------------------------------------------------------------------------------------------
    def _run_cycle(self):
        """
        Custom implementation of a single process cycle. To be redefined.

        Returns
        -------
        success : bool
            True on success. False otherwise.
        error : bool
            True on error. False otherwise.
        adapted : bool
            True, if ml model adapted something in this cycle. False otherwise.

        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def get_cycle_id(self):
        """
        Returns current cycle id.
        """

        return self._cycle_id


## -------------------------------------------------------------------------------------------------
    def run(self, 
            p_term_on_success:bool=True,        
            p_term_on_error:bool=True,          
            p_term_on_timeout:bool=False ):    
        """
        Runs the scenario as a sequence of single process steps until a terminating event occures.

        Parameters
        ----------
        p_term_on_success : bool
            If True, the run terminates on success. Default = True.
        p_term_on_error : bool
            If True, the run terminates on error. Default = True.
        p_term_on_timeout : bool
            If True, the run terminates on timeout. Default = False.

        Returns
        -------
        success : bool
            True on success. False otherwise.
        error : bool
            True on error. False otherwise.
        timeout : bool
            True on timeout. False otherwise.
        cycle_limit : bool
            True, if cycle limit has reached. False otherwise.
        adapted : bool
            True, if ml model adapted something. False otherwise.
        num_cycles: int
            Number of cycles.

        """

        self._cycle_id  = 0
        adapted         = False
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), 'Start of processing')

        while True:
            success, error, timeout, limit, adapted_cycle = self.run_cycle()
            adapted = adapted or adapted_cycle
            if p_term_on_success and success: break
            if p_term_on_error and error: break
            if p_term_on_timeout and timeout: break
            if limit: break

        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), 'End of processing')

        return success, error, timeout, limit, adapted, self._cycle_id





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class TrainingResults (Log, Saveable):
    """
    Results of a training (see class Training).

    Parameters
    ----------
    p_scenario : Scenario
        Related scenario.
    p_run : int
        Run id.
    p_cycle_id : int
        Id of first cycle of this run.
    p_path : str
        Optional estination path to store the results.
    p_logging
        Log level (see constants of class Log). Default: Log.C_LOG_ALL

    """

    C_TYPE      = 'Results '

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_scenario:Scenario, p_run, p_cycle_id, p_path=None, p_logging=Log.C_LOG_WE):
        self.scenario           = p_scenario
        self.run                = p_run
        self.ts_start           = datetime.now()
        self.ts_end             = None
        self.duration           = 0
        self.cycle_id_start     = p_cycle_id
        self.cycle_id_end       = 0
        self.num_cycles         = 0
        self.num_cycles_train   = 0
        self.num_cycles_eval    = 0
        self.num_adaptations    = 0
        self.highscore          = None
        self.path               = p_path

        self.custom_results     = []

        Log.__init__(self, p_logging=p_logging)


## -------------------------------------------------------------------------------------------------
    def add_custom_result(self, p_name, p_value):
        self.custom_results.append([p_name, p_value])


## -------------------------------------------------------------------------------------------------
    def close(self):
        self.ts_end             = datetime.now()
        self.duration           = self.ts_end - self.ts_start
        self.cycle_id_end       = self.cycle_id_start + self.num_cycles -1


## -------------------------------------------------------------------------------------------------
    def log_results(self):
        self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
        self.log(self.C_LOG_TYPE_W, '-- Training Results of run', self.run)
        self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
        self._log_results()
        self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
        self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n')


## -------------------------------------------------------------------------------------------------
    def _log_results(self):
        self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
        self.log(self.C_LOG_TYPE_W, '-- Scenario          :', self.scenario.C_TYPE, self.scenario.C_NAME)
        self.log(self.C_LOG_TYPE_W, '-- Model             :', self.scenario.get_model().C_TYPE, self.scenario.get_model().C_NAME)
        self.log(self.C_LOG_TYPE_W, '-- Start time stamp  :', self.ts_start)
        self.log(self.C_LOG_TYPE_W, '-- End time stamp    :', self.ts_end)
        self.log(self.C_LOG_TYPE_W, '-- Duration          :', self.duration)
        self.log(self.C_LOG_TYPE_W, '-- Start cycle id    :', self.cycle_id_start)
        self.log(self.C_LOG_TYPE_W, '-- End cycle id      :', self.cycle_id_end)
        self.log(self.C_LOG_TYPE_W, '-- Training cycles   :', self.num_cycles_train)
        self.log(self.C_LOG_TYPE_W, '-- Evaluation cycles :', self.num_cycles_eval)
        self.log(self.C_LOG_TYPE_W, '-- Adaptations       :', self.num_adaptations)
        self.log(self.C_LOG_TYPE_W, '-- High score        :', self.highscore)
        if self.path is not None:
            self.log(self.C_LOG_TYPE_W, '-- Results stored in : "' + self.path +'"')


## -------------------------------------------------------------------------------------------------
    def _save_line(self, p_file, p_name, p_value):
        value = p_value
        if value is None: value = '-'
        p_file.write(p_name + '\t' + str(value) + '\n')


## -------------------------------------------------------------------------------------------------
    def save(self, p_path, p_filename='summary.csv') -> bool:
        """
        Saves a training summary in the given path.

        Parameters
        ----------
        p_path : str
            Destination folder
        p_filename  :string
            Name of summary file. Default = 'summary.csv'

        Returns
        -------
        success : bool
            True, if summary file was created successfully. False otherwise.

        """

        filename = p_path + os.sep + p_filename
        filename.replace(os.sep + os.sep, os.sep)

        file = open(filename, 'wt')
        if file is None: return False

        self._save_line(file, 'Scenario', '"' + self.scenario.C_TYPE + ' ' + self.scenario.C_NAME + '"')
        self._save_line(file, 'Model', '"' + self.scenario.get_model().C_TYPE + ' ' + self.scenario.get_model().C_NAME + '"')
        self._save_line(file, 'Run', str(self.run))
        self._save_line(file, 'Start', self.ts_start)
        self._save_line(file, 'End', self.ts_end)
        self._save_line(file, 'Duration', self.duration)
        self._save_line(file, 'Start cycle id', self.cycle_id_start)
        self._save_line(file, 'End cycle id', self.cycle_id_end)
        self._save_line(file, 'Training cycles', self.num_cycles_train)
        self._save_line(file, 'Evaluation cycles', self.num_cycles_eval)
        self._save_line(file, 'Adaptations', self.num_adaptations)
        self._save_line(file, 'Highscore', self.highscore)
        self._save_line(file, 'Path', '"' + str(self.path) + '"')

        for name, value in self.custom_results:
            self._save_line(file, name, value)

        file.close()
        return True





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class HyperParamTuner (Log, Saveable):
    """
    Template class for hyperparameter tuning (HPT).
    """

    C_TYPE      = 'HyperParam Tuner'
    C_NAME      = '????'
    C_VAR_TRIAL = 'Trial'
    C_VAR_SCORE = 'Highscore'

## -------------------------------------------------------------------------------------------------
    def maximize(self, p_training_cls, p_num_trials, p_root_path, **p_training_param ) -> TrainingResults:
        """
        ...

        Parameters
        ----------
        p_training_cls
            Training class to be instantiated/executed 
        p_num_trials : int    
            Number of trials
        p_num_trials : str    
            Root path of the training class
        p_training_param : dictionary
            Training parameters

        Returns
        -------
        results : TrainingResults
            Training results of the best tuned model (see class TrainingResults).

        """

        self._training_cls      = p_training_cls
        self._num_trials        = p_num_trials
        self._root_path         = p_root_path
        self._training_param    = p_training_param
        self.HPDataStoring      = None
        self.variables          = [self.C_VAR_TRIAL, self.C_VAR_SCORE]

        return self._maximize()


## -------------------------------------------------------------------------------------------------
    def _maximize(self) -> TrainingResults:
        raise NotImplementedError
    
## -------------------------------------------------------------------------------------------------
    def _save_line(self, p_file, p_name, p_value):
        value = p_value
        if value is None: value = '-'
        p_file.write(p_name + '\t' + str(value) + '\n')


## -------------------------------------------------------------------------------------------------
    def save(self, p_param, p_result, p_filename='best_parameters.csv') -> bool:
        """
        Saves the best result of the hyperparameter tuning in the root path.

        Parameters
        ----------
        p_param : dict
            A dictionary that consists of list of best parameters
        p_result : float
            Highest score
        p_filename  :str
            Name of summary file. Default = 'best_parameters.csv'

        Returns
        -------
        success : bool
            True, if summary file was created successfully. False otherwise.

        """

        filename = self._root_path + os.sep + p_filename
        filename.replace(os.sep + os.sep, os.sep)

        file = open(filename, 'wt')
        if file is None: return False
  
        self._save_line(file, 'Tuner', '"' + self.C_NAME + '"')    
        self._save_line(file, 'Number of evaluations', self._num_trials)     
        self._save_line(file, 'Highest Score', p_result)
        for key in p_param:
            self._save_line(file, key, p_param[key])

        file.close()
        
        try:
            self.HPDataStoring.save_data(self._root_path, 'tuning_summary', '\t')
            return True
        except:
            return False
        




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Training (Log):
    """
    Template class for a ML training and hyperparameter tuning.

    Parameters
    ----------
    p_scenario_cls 
        Name of ML scenario class, compatible to/inherited from class Scenario.
    p_cycle_limit : int
        Maximum number of training cycles (0=no limit). Default = 0.
    p_adaptation_limit : int
        Maximum number of adaptations (0=no limit). Default = 0.
    p_hpt : HyperParamTuner
        Optional hyperparameter tuner (see class HyperParamTuner). Default = None.
    p_hpt_trials : int
        Optional number of hyperparameter tuning trials. Default = 0.        
    p_path : str
        Optional destination path to store training data. Default = None.
    p_visualize : bool
        Boolean switch for env/agent visualisation. Default = False
    p_logging
        Log level (see constants of class Log). Default = Log.C_LOG_WE.

    """

    C_TYPE          = 'Training'
    C_NAME          = '????'

    C_CLS_RESULTS   = TrainingResults

    C_MODE_TRAIN    = 0 
    C_MODE_EVAL     = 1

    C_LOG_SEPARATOR = '------------------------------------------------------------------------------'

## -------------------------------------------------------------------------------------------------
    def __init__(self, **p_kwargs):
 
        # 1 Check and completion of parameters
        self._kwargs = p_kwargs.copy()

        # 1.1 Mandatory parameter p_scenario_cls
        try:
            scenario_cls = self._kwargs['p_scenario_cls']
        except:
            raise ParamError('Mandatory parameter p_scenario_cls not supplied')

        # 1.2 Optional parameter p_cycle_limit
        try:
            self._cycle_limit = self._kwargs['p_cycle_limit']
        except:
            self._cycle_limit = 0
            self._kwargs['p_cycle_limit'] = self._cycle_limit

        # 1.3 Optional parameter p_adaptation_limit
        try:
            self._adaptation_limit = self._kwargs['p_adaptation_limit']
        except:
            self._adaptation_limit = 0
            self._kwargs['p_adaptation_limit'] = self._adaptation_limit

        # 1.4 Optional parameter p_hpt
        try:
            self._hpt = self._kwargs['p_hpt']
        except:
            self._hpt = None
            self._kwargs['p_hpt'] = self._hpt

        # 1.5 Optional parameter p_hpt_trials
        try:
            self._hpt_trials = self._kwargs['p_hpt_trials']
        except:
            self._hpt_trials = 0
            self._kwargs['p_hpt_trials'] = self._hpt_trials 

        if ( self._hpt is not None ) and ( self._hpt_trials <= 0 ):
            raise ParamError('Please supply parameter p_hpt_trials with a number >0')

        # 1.6 Optional file path
        try:
            path = self._kwargs['p_path']
        except:
            path = None
            self._kwargs['p_path'] = path

        # 1.7 Optional parameter for visualization
        try:
            visualize = self._kwargs['p_visualize']
        except:
            visualize = False
            self._kwargs['p_visualize'] = visualize

        # 1.8 Optional log level
        try:
            logging = self._kwargs['p_logging']
        except:
            logging = Log.C_LOG_WE
            self._kwargs['p_logging'] = logging

        # 1.9 Optional environment mode
        try:
            env_mode = self._kwargs['p_env_mode']
        except:
            env_mode = Mode.C_MODE_SIM
            self._kwargs['p_env_mode'] = env_mode


        # 2 Initialization
        super().__init__(p_logging=logging)

        self._current_run       = 0
        self._new_run           = True
        self._results           = None
        self._root_path         = self._gen_root_path(path)
        self._current_path      = None
        self._scenario          = None
        self._mode              = self.C_MODE_TRAIN


        # 3 Setup scenario
        if self._hpt is None:
            try:
                self._scenario = scenario_cls( p_mode=env_mode, 
                                               p_ada=True,
                                               p_cycle_limit=self._cycle_limit,
                                               p_visualize=visualize,
                                               p_logging=logging )
            except:
                raise ParamError('Par p_scenario_cls: class "' + scenario_cls.__name__ + '" not compatible')


## -------------------------------------------------------------------------------------------------
    def get_scenario(self) -> Scenario:
        return self._scenario


## -------------------------------------------------------------------------------------------------
    def _gen_root_path(self, p_path) -> str:
        if p_path is None: return None

        now         = datetime.now()
        ts          = '%04d-%02d-%02d  %02d.%02d.%02d' % (now.year, now.month, now.day, now.hour, now.minute, now.second)
        root_path   = p_path + str(os.sep) + ts + ' ' + self.C_TYPE + ' ' + self.C_NAME
        root_path.replace(str(os.sep) + str(os.sep), str(os.sep))
        os.mkdir(root_path)
        return root_path


## -------------------------------------------------------------------------------------------------
    def _gen_current_path(self, p_root_path, p_run) -> str:
        if p_root_path is None: return None
        if self._hpt is None: return self._root_path

        now             = datetime.now()
        ts              = '%04d-%02d-%02d  %02d.%02d.%02d' % (now.year, now.month, now.day, now.hour, now.minute, now.second)
        current_path    = p_root_path + os.sep + ts + ' Run #' + str(p_run)
        current_path.replace(os.sep + os.sep, os.sep)
        os.mkdir(current_path)
        return current_path


## -------------------------------------------------------------------------------------------------
    def _init_results(self) -> TrainingResults:
        return self.C_CLS_RESULTS(self._scenario, 
                                  self._current_run, 
                                  self._scenario.get_cycle_id(), 
                                  p_path=self._current_path,
                                  p_logging=self._level)


## -------------------------------------------------------------------------------------------------
    def _close_results(self, p_results:TrainingResults):
        p_results.close()
        if self._current_path is not None: p_results.save(self._current_path)


## -------------------------------------------------------------------------------------------------
    def run_cycle(self) -> bool:
        """
        Runs a single training cycle.
        
        Returns
        -------
        termination_event : bool
            True, if training run has finished. False otherwise.

        """

        # 1 Intro
        if self._new_run:
            # 1.1 Start of new training run
            self._current_path  = self._gen_current_path(self._root_path, self._current_run)
            self._results       = self._init_results()
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, '-- Training run', self._current_run, 'started...')
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR, '\n')
            self._new_run = False
            

        # 2 Run a single training cycle
        mode         = self._mode
        run_finished = self._run_cycle()

        if mode == self.C_MODE_TRAIN:
            self._results.num_cycles_train += 1
            self._results.num_cycles += 1
        elif mode == self.C_MODE_EVAL:
            self._results.num_cycles_eval += 1


        # 3 Assess results
        if ( self._cycle_limit > 0 ) and ( self._results.num_cycles_train >= self._cycle_limit ):
            # 3.1 Cycle limit reached
            self.log(self.C_LOG_TYPE_W, 'Training cycle limit', self._cycle_limit, 'reached')
            run_finished = True

        if run_finished:
            # 3.2 Training run finished
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, '-- Training run', self._current_run, 'finished')
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, self.C_LOG_SEPARATOR, '\n')

            self._scenario.get_model().save(self._current_path, 'trained model.pkl')
            self._close_results(self._results)
            self._results.log_results()

            self._current_path = None
            self._new_run       = True

        return run_finished
        

## -------------------------------------------------------------------------------------------------
    def _run_cycle(self) -> bool:
        """
        Single custom trainig cycle to be redefined. Custom training results can be added to using
        self._results.add_custom_result(p_name, p_value).

        Returns
        -------
        bool
            True, if training has finished. False otherwise.

        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def run(self) -> TrainingResults:
        """
        Runs a training and returns the results of the best trained/tuned agent.

        Returns
        -------
        TrainingResults
            Object with training results.

        """

        if self._hpt is None:
            # 1 Training without hyperparameter tuning
            self.log(self.C_LOG_TYPE_I, 'Training started (without hyperparameter tuning)')
            self._results = self._run()

        else:
            # 2 Training with hyperparameter tuning
            self.log(self.C_LOG_TYPE_I, 'Training started (with hyperparameter tuning)')
            training_param  = self._kwargs.copy()
            training_param.pop('p_hpt')
            training_param.pop('p_hpt_trials')
            self._results = self._hpt.maximize(p_training_cls=self.__class__, p_num_trials=self._hpt_trials, p_root_path=self._root_path, p_training_param=training_param)

        self.log(self.C_LOG_TYPE_I, 'Training completed')
        return self.get_results()


## -------------------------------------------------------------------------------------------------
    def _run(self) -> TrainingResults:
        self._new_run = True
        while not self.run_cycle(): pass
        return self.get_results()


## -------------------------------------------------------------------------------------------------
    def get_results(self) -> TrainingResults:
        return self._results


#!/usr/bin/env python

import rospy
from gazebo_msgs.srv import SpawnModel, SpawnModelRequest, SpawnModelResponse
from copy import deepcopy
from tf.transformations import quaternion_from_euler

sdf_cube = """<?xml version="1.0" ?>
<sdf version="1.4">
  <model name="MODELNAME">
    <static>0</static>
    <link name="link">
      <inertial>
        <mass>1.0</mass>
        <inertia>
          <ixx>0.01</ixx>
          <ixy>0.0</ixy>
          <ixz>0.0</ixz>
          <iyy>0.01</iyy>
          <iyz>0.0</iyz>
          <izz>0.01</izz>
        </inertia>
      </inertial>
      <collision name="stairs_collision0">
        <pose>0 0 0 0 0 0</pose>
        <geometry>
          <box>
            <size>SIZEXYZ</size>
          </box>
        </geometry>
        <surface>
          <bounce />
          <friction>
            <ode>
              <mu>1.0</mu>
              <mu2>1.0</mu2>
            </ode>
          </friction>
          <contact>
            <ode>
              <kp>10000000.0</kp>
              <kd>1.0</kd>
              <min_depth>0.0</min_depth>
              <max_vel>0.0</max_vel>
            </ode>
          </contact>
        </surface>
      </collision>
      <visual name="stairs_visual0">
        <pose>0 0 0 0 0 0</pose>
        <geometry>
          <box>
            <size>SIZEXYZ</size>
          </box>
        </geometry>
        <material>
          <script>
            <uri>file://media/materials/scripts/gazebo.material</uri>
            <name>Gazebo/Wood</name>
          </script>
        </material>
      </visual>
      <velocity_decay>
        <linear>0.000000</linear>
        <angular>0.000000</angular>
      </velocity_decay>
      <self_collide>0</self_collide>
      <kinematic>0</kinematic>
      <gravity>1</gravity>
    </link>
  </model>
</sdf>
"""


def create_cube_request(modelname, px, py, pz, rr, rp, ry, sx, sy, sz):
    """Create a SpawnModelRequest with the parameters of the cube given.
    modelname: name of the model for gazebo
    px py pz: position of the cube (and it's collision cube)
    rr rp ry: rotation (roll, pitch, yaw) of the model
    sx sy sz: size of the cube"""
    cube = deepcopy(sdf_cube)
    # Replace size of model
    size_str = str(round(sx, 3)) + " " + \
        str(round(sy, 3)) + " " + str(round(sz, 3))
    cube = cube.replace('SIZEXYZ', size_str)
    # Replace modelname
    cube = cube.replace('MODELNAME', str(modelname))

    req = SpawnModelRequest()
    req.model_name = modelname
    req.model_xml = cube
    req.initial_pose.position.x = px
    req.initial_pose.position.y = py
    req.initial_pose.position.z = pz

    q = quaternion_from_euler(rr, rp, ry)
    req.initial_pose.orientation.x = q[0]
    req.initial_pose.orientation.y = q[1]
    req.initial_pose.orientation.z = q[2]
    req.initial_pose.orientation.w = q[3]

    return req


if __name__ == '__main__':
    rospy.init_node('spawn_models')
    spawn_srv = rospy.ServiceProxy('/gazebo/spawn_sdf_model', SpawnModel)
    rospy.loginfo("Waiting for /gazebo/spawn_sdf_model service...")
    spawn_srv.wait_for_service()
    rospy.loginfo("Connected to service!")

    # Spawn object 1
    rospy.loginfo("Spawning cube1")
    req1 = create_cube_request("cube1",
                              0.0, 0.0, 0.51,  # position
                              0.0, 0.0, 0.0,  # rotation
                              1.0, 1.0, 1.0)  # size
    spawn_srv.call(req1)
    rospy.sleep(1.0)

    # Spawn object 2
    rospy.loginfo("Spawning cube2")
    req2 = create_cube_request("cube2",
                              0.0, 1.1, 0.41,  # position
                              0.0, 0.0, 0.0,  # rotation
                              0.8, 0.8, 0.8)  # size
    spawn_srv.call(req2)
    rospy.sleep(1.0)

    # Spawn object 3
    rospy.loginfo("Spawning cube3")
    req3 = create_cube_request("cube3",
                              0.0, 2.1, 0.41,  # position
                              0.0, 0.0, 0.0,  # rotation
                              0.4, 0.4, 0.4)  # size
    spawn_srv.call(req3)
    rospy.sleep(1.0)



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_011_train_ur5_environment_with_wrapped_sb3_policy.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-11-18  0.0.0     MRD      Creation
## -- 2021-11-18  1.0.0     MRD      Initial Release
## -- 2021-12-07  1.0.1     DA       Refactoring
## -- 2022-02-11  1.1.0     DA       Special derivate for publication
## -- 2022-05-23  1.2.0     MRD      Add visualize toggle on UR5JointControl for gazebo GUI
## -- 2022-06-06  1.2.1     MRD      Add real connection option
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.2.1 (2022-06-06)

This module shows how to use SB3 wrapper to train UR5 robot (derivate for paper).
"""

from mlpro.rl.models import *
from mlpro.rl.pool.envs.ur5jointcontrol import UR5JointControl
from stable_baselines3 import PPO
from mlpro.wrappers.sb3 import WrPolicySB32MLPro
from pathlib import Path


# 1 Implement your own RL scenario
class ScenarioUR5A2C(RLScenario):
    C_NAME = 'Matrix'

    def _setup(self, p_mode, p_ada, p_logging):
        # 1.1 Setup environment
        self._env = UR5JointControl(
            p_build=True, 
            p_real=p_mode,
            # p_robot_ip="172.19.10.41",
            # p_reverse_ip="172.19.10.170", 
            p_visualize=self._visualize, 
            p_logging=p_logging)

        policy_sb3 = PPO(
            policy="MlpPolicy",
            n_steps=20,
            env=None,
            _init_setup_model=False,
            device="cpu",
            seed=1)

        policy_wrapped = WrPolicySB32MLPro(
            p_sb3_policy=policy_sb3,
            p_cycle_limit=self._cycle_limit,
            p_observation_space=self._env.get_state_space(),
            p_action_space=self._env.get_action_space(),
            p_ada=p_ada,
            p_logging=p_logging)

        # 1.2 Setup standard single-agent with own policy
        return Agent(
            p_policy=policy_wrapped,
            p_envmodel=None,
            p_name='Smith',
            p_ada=p_ada,
            p_logging=p_logging
        )


# 2 Train agent in scenario
now = datetime.now()

training = RLTraining(
    p_scenario_cls=ScenarioUR5A2C,
    p_env_mode=Mode.C_MODE_REAL,
    p_cycle_limit=5500,
    p_cycles_per_epi_limit=-1,
    p_collect_states=True,
    p_collect_actions=True,
    p_collect_rewards=True,
    p_collect_training=True,
    p_visualize=True,
    p_path=str(Path.home()),
    p_logging=Log.C_LOG_WE)

training.run()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl.pool.envs
## -- Module  : gridworld
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-06  0.0.0     WB       Creation
## -- 2021-09-09  1.0.0     WB       Release of first version
## -- 2021-09-11  1.0.1     MRD      Fix compability with mlpro structure
## -- 2021-09-11  1.0.1     MRD      Change Header information to match our new library name
## -- 2021-09-13  1.0.2     WB       Fix on simulate reaction      
## -- 2021-09-30  1.0.3     SY       State-space and action-space improvement     
## -- 2021-10-05  1.0.4     SY       Update following new attributes done and broken in State
## -- 2021-11-15  1.0.5     DA       Refactoring
## -- 2021-12-03  1.0.6     DA       Refactoring
## -- 2021-12-19  1.0.7     DA       Replaced 'done' by 'success'
## -- 2021-12-21  1.0.8     DA       Class GridWorld: renamed method reset() to _reset()
## -- 2022-02-25  1.0.9     SY       Refactoring due to auto generated ID in class Dimension
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.9 (2021-02-25)

This module provides an environment of customizable Gridworld.
"""

from mlpro.rl.models import *
from mlpro.gt.models import *
import numpy as np
         
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class GridWorld (Environment):
    """
    Custom environment of an n-D grid world where the agent 
    has to go to a random target.
    """
    C_NAME          = 'GridWorld'
    C_LATENCY       = timedelta(0,1,0)
    C_INFINITY      = np.finfo(np.float32).max  
    C_REWARD_TYPE   = Reward.C_TYPE_OVERALL
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, p_logging=True, grid_size=(8,8),
                random_start_position=True,
                random_goal_position=True, max_step=10):
        """
        Parameters:
            p_logging               Boolean switch for logging
            grid_size               Dimension of the grid world
            random_start_position   Randomize start position
            random_goal_position    Randomize goal position
            max_step                Maximum step per episode
        """
        
                
        self.grid_size = np.array(grid_size)
        self.random_start_position = random_start_position
        self.random_goal_position = random_goal_position

        self.agent_pos = np.array([np.random.randint(0,border-1) 
                            if self.random_start_position 
                            else 0 for border in self.grid_size])

        self.goal_pos = np.array([np.random.randint(0,border-1) 
                            if self.random_goal_position 
                            else border-1 for border in self.grid_size])
        self.max_step = max_step
        self.num_step = 0
        
        super().__init__(p_mode=Environment.C_MODE_SIM, p_logging=p_logging)
        self._state_space, self._action_space = self._setup_spaces()

        self.reset()


## -------------------------------------------------------------------------------------------------
    @staticmethod
    def setup_spaces():
        return None, None

        
## -------------------------------------------------------------------------------------------------
    def _setup_spaces(self):

        state_space     = ESpace()
        action_space    = ESpace()
        data            = 1

        for size in self.grid_size:
            data *= size
            
        for i in range(data):
            state_space.add_dim(Dimension( p_name_short=str(i), p_base_set=Dimension.C_BASE_SET_Z, p_boundaries=[0,3]))
        
        for i in range(len(self.grid_size)):
            action_space.add_dim(Dimension( p_name_short=str(i), p_base_set=Dimension.C_BASE_SET_Z, p_boundaries=[-self.grid_size[i], self.grid_size[i]]))

        return state_space, action_space

    
## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None) -> None:
        """
        To reset environment
        """
        random.seed(p_seed)
        self.agent_pos = np.array([np.random.randint(0,border-1) 
                            if self.random_start_position 
                            else 0 for border in self.grid_size])

        self.goal_pos = np.array([np.random.randint(0,border) 
                            if self.random_goal_position 
                            else border-1 for border in self.grid_size])
        self.num_step = 0
        self._state = self.get_state()
        

## -------------------------------------------------------------------------------------------------
    def get_state(self):
        obs = np.zeros(self.grid_size, dtype=np.float32)
        if np.allclose(self.agent_pos, self.goal_pos):
            obs[tuple(self.agent_pos)] = 3
        else:
            obs[tuple(self.agent_pos)] = 1
            obs[tuple(self.goal_pos)] = 2
        state = State(self._state_space)
        state.set_values(obs.flatten())
        return state
        

## -------------------------------------------------------------------------------------------------
    def _simulate_reaction(self, p_state:State, p_action:Action) -> State:
        self.agent_pos += np.array(p_action.get_sorted_values()).astype(int)
        self.agent_pos = np.clip(self.agent_pos, 0, self.grid_size-1)
        
        self.num_step += 1
        euclidean_distance = np.linalg.norm(self.goal_pos-self.agent_pos)
        if euclidean_distance == 0:
            self._state.set_success(True)
        else:
            self._state.set_success(False)
        self._state = self.get_state()
        return self._state
        

## -------------------------------------------------------------------------------------------------
    def _compute_reward(self, p_state_old:State, p_state_new:State) -> Reward:
        reward = Reward(self.C_REWARD_TYPE)
        rew = 1
        euclidean_distance = np.linalg.norm(self.goal_pos-self.agent_pos).item()
        if euclidean_distance !=0:
            rew = 1/euclidean_distance
        if self.num_step >= self.max_step:
            rew -= self.max_step
        
        reward.set_overall_reward(rew)
        return reward


## -------------------------------------------------------------------------------------------------
    def _compute_success(self, p_state:State) -> bool:
        if self.num_step >= self.max_step:
            return True
        elif self.get_success() == True:
            return True

        return False


## -------------------------------------------------------------------------------------------------
    def _compute_broken(self, p_state:State) -> bool:
        return False
        

## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        pass


## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        pass


## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : test_pool_policies
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-08  1.0.0     MRD      Creation
## -- 2022-05-19  1.0.1     SY       Add RandomGenerator policy
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.0 (2021-12-08)

Unit test classes for policies.
"""


import pytest
import random
import gym
import numpy as np
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro
from mlpro.rl.pool.policies.dummy import MyDummyPolicy
from mlpro.rl.pool.policies.randomgenerator import RandomGenerator

## Instructions
# Please Include your own policy implementation class on the test list parameter


## -------------------------------------------------------------------------------------------------
@pytest.mark.parametrize("policy_cls", [MyDummyPolicy, RandomGenerator])
def test_pool_policies(policy_cls):
    class MyScenario (RLScenario):

        C_NAME      = 'Matrix'

        def _setup(self, p_mode, p_ada, p_logging):
            # 1 Setup environment
            gym_env     = gym.make('CartPole-v1')
            self._env   = WrEnvGYM2MLPro(gym_env, p_logging=False) 

            # 2 Setup and return standard single-agent with own policy
            return Agent(
                p_policy=policy_cls(
                    p_observation_space=self._env.get_state_space(),
                    p_action_space=self._env.get_action_space(),
                    p_buffer_size=10,
                    p_ada=p_ada,
                    p_logging=p_logging
                ),    
                p_envmodel=None,
                p_name='Smith',
                p_ada=p_ada,
                p_logging=p_logging
            )

    # 2.4 Create and run training object
    training = RLTraining(
            p_scenario_cls=MyScenario,
            p_cycle_limit=100,
            p_max_adaptations=0,
            p_max_stagnations=0,
            p_logging=False
    )

    training.run()


## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro.bf
## -- Module  : plot
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-10-06  1.0.0     DA       Creation and transfer of classes DataPlotting, Plottable from 
## --                                mlpro.bf.various
## -- 2021-10-25  1.0.1     SY       Improve get_plots() functionality, enable episodic plots
## -- 2021-12-10  1.0.2     SY       Add errors and exceptions, if p_printing is None.
## --                                Clean code assurance.
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.2 (2021-12-10)

This module provides various classes related to data plotting.
"""

import numpy as np
import matplotlib.pyplot as plt
import os
from mlpro.bf.various import LoadSave
from mlpro.bf.data import DataStoring
import statistics


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Plottable:
    """
    Property class that inherits the ability to be plottable.
    The constructor initializes the plot. Optionally the plot itself will be embedded in a matplotlib figure.

    Parameters
    ----------
    p_figure : TYPE, optional
        Optional MatPlotLib host figure, where the plot shall be embedded. The default is None.
    """

    ## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        pass

    ## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        """
        Updates the plot.
        """

        pass


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DataPlotting(LoadSave):
    """
    This class provides a functionality to plot the stored values of variables.

    Parameters
    ----------
    p_data : DataStoring
        Data object with stored variables values.
    p_type : str, optional
        Type of plot. The default is C_PLOT_TYPE_EP.
    p_window : int, optional
        Moving average parameter. The default is 100.
    p_showing : Bool, optional
        Showing graphs after they are generated. The default is True.
    p_printing : dict, optional
        Additional important parameters for plotting.
        [0] = Bool : Whether the stored values is plotted.
        [1] = Float : Min. value on graph.
        [2] = Float : Max. value on graph. Set to -1, if you want to set min/max value according to the stored values.
        Example = {"p_variable_1" : [True,0,-1],
                   "p_variable_2" : [True,-0.5,10]}.
        The default is None.
    p_figsize : int, optional
        Frame size. The default is (7,7).
    p_color : str, optional
        Line colors. The default is "darkblue".
        
    Attributes
    ----------
    C_PLOT_TYPE_CY : str
        one of the plotting types, which plot the graph with multiple lines according to the number of frames.
    C_PLOT_TYPE_EP : str
        one of the plotting types, which plot the graph everything in one line regardless the number of frames.
    C_PLOT_TYPE_EP_M : str
        one of the plotting types, which plot only the mean value of each variable for each frame.
        
    """

    C_PLOT_TYPE_CY = 'Cyclic'
    C_PLOT_TYPE_EP = 'Episodic'
    C_PLOT_TYPE_EP_M = 'Episodic Mean'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_data: DataStoring, p_type=C_PLOT_TYPE_EP, p_window=100,
                 p_showing=True, p_printing=None, p_figsize=(7, 7), p_color="darkblue"):
        self.data = p_data
        self.type = p_type
        self.window = p_window
        self.showing = p_showing
        self.plots = [[], []]
        self.printing = p_printing
        self.figsize = p_figsize
        self.color = p_color

    ## -------------------------------------------------------------------------------------------------
    def get_plots(self):
        """
        A function to plot data.
        """

        if self.type == 'Cyclic':
            self.plots_type_cy()
        elif self.type == 'Episodic':
            self.plots_type_ep()
        elif self.type == 'Episodic Mean':
            self.plots_type_ep_mean()

    ## -------------------------------------------------------------------------------------------------
    def plots_type_cy(self):
        """
        A function to plot data per cycle.
        """

        for name in self.data.names:
            maxval = 0
            try:
                if self.printing[name][0]:
                    fig = plt.figure(figsize=self.figsize)
                    lines = []
                    label = []
                    plt.title(name)
                    plt.grid(True, which="both", axis="both")
                    for fr in range(len(self.data.memory_dict[name])):
                        fr_id = self.data.frame_id[name][fr]
                        lines += plt.plot(self.moving_mean(self.data.get_values(name, fr_id), self.window),
                                          color=self.color, alpha=(fr + 1.0) / (len(self.data.memory_dict[name]) + 1))
                        if self.printing[name][2] == -1:
                            maxval = max(max(self.data.get_values(name, fr_id)), maxval)
                        else:
                            maxval = self.printing[name][2]
                        label.append("%s" % fr_id)
                    plt.ylim(self.printing[name][1], maxval)
                    plt.xlabel("cycles")
                    plt.legend(label, bbox_to_anchor=(1, 0.5), loc="center left")
                    self.plots[0].append(name)
                    self.plots[1].append(fig)
                    if self.showing:
                        plt.show()
                    else:
                        plt.close(fig)
            except:
                pass

    ## -------------------------------------------------------------------------------------------------
    def plots_type_ep(self):
        """
        A function to plot data per frame by extending the cyclic plots in one plot.
        """

        for name in self.data.names:
            maxval = 0
            try:
                if self.printing[name][0]:
                    fig = plt.figure(figsize=self.figsize)
                    lines = []
                    data = []
                    plt.title(name)
                    plt.grid(True, which="both", axis="both")
                    for fr in range(len(self.data.memory_dict[name])):
                        fr_id = self.data.frame_id[name][fr]
                        data.extend(self.data.get_values(name, fr_id))
                        if self.printing[name][2] == -1:
                            maxval = max(max(self.data.get_values(name, fr_id)), maxval)
                        else:
                            maxval = self.printing[name][2]
                    lines += plt.plot(self.moving_mean(data[:], self.window), color=self.color)
                    plt.ylim(self.printing[name][1], maxval)
                    plt.xlabel("continuous cycles")
                    self.plots[0].append(name)
                    self.plots[1].append(fig)
                    if self.showing:
                        plt.show()
                    else:
                        plt.close(fig)
            except:
                pass

    ## -------------------------------------------------------------------------------------------------
    def plots_type_ep_mean(self):
        """
        A function to plot data per frame according to its mean value.
        """

        for name in self.data.names:
            maxval = 0
            try:
                if self.printing[name][0]:
                    fig = plt.figure(figsize=self.figsize)
                    lines = []
                    data = []
                    plt.title(name)
                    plt.grid(True, which="both", axis="both")
                    for fr in range(len(self.data.memory_dict[name])):
                        fr_id = self.data.frame_id[name][fr]
                        data.extend([statistics.mean(self.data.get_values(name, fr_id))])
                    if self.printing[name][2] == -1:
                        maxval = max(max(data[:]), maxval)
                    else:
                        maxval = self.printing[name][2]
                    lines += plt.plot(self.moving_mean(data[:], self.window), color=self.color)
                    plt.ylim(self.printing[name][1], maxval)
                    plt.xlabel("episodes")
                    self.plots[0].append(name)
                    self.plots[1].append(fig)
                    if self.showing:
                        plt.show()
                    else:
                        plt.close(fig)
            except:
                pass

    ## -------------------------------------------------------------------------------------------------
    def moving_mean(self, p_inputs, p_window):
        """
        This method creates a series of averages of different subsets of the full data set.

        Parameters
        ----------
        p_inputs : list of floats
            input dataset.
        p_window : int
            moving average parameter.

        Returns
        -------
        outputs : list of floats
            transformed data set.

        """

        inputs = np.array(p_inputs)
        outputs = np.zeros_like(inputs)
        if len(inputs.shape) == 1:
            outputs = np.convolve(inputs, np.ones((p_window,)) / p_window, mode='same')
        else:
            for col in range(inputs.shape[1]):
                outputs[:, col] = np.convolve(inputs[:, col], np.ones((p_window,)) / p_window, mode='same')
        return outputs

    ## -------------------------------------------------------------------------------------------------
    def save_plots(self, p_path, p_format, p_dpi_mul=1):
        """
        This method is used to save generated plots.

        Parameters
        ----------
        p_path : str
            Path where file will be saved.
        p_format : str
            Format of the saved file.
            Options: 'eps', 'jpg', 'png', 'pdf', 'svg'.
        p_dpi_mul : int, optional
            Saving plots parameter. The default is 1.

        Returns
        -------
        bool
            True, if plots where saved successfully. False otherwise..

        """

        num_plots = len(self.plots[0])
        if num_plots == 0: return False

        try:
            if not os.path.exists(p_path):
                os.makedirs(p_path)
            for idx in range(num_plots):
                self.plots[1][idx].savefig(p_path + os.sep + self.plots[0][idx] + "." + p_format, dpi=500 * p_dpi_mul,
                                           bbox_inches='tight')
            return True
        except:
            return False



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_007_train_wrapped_SB3_policy.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-29  0.0.0     MRD      Creation
## -- 2021-10-07  1.0.0     MRD      Released first version
## -- 2021-10-08  1.0.1     DA       Take over the cycle limit from the environment
## -- 2021-10-18  1.0.2     DA       Refactoring
## -- 2021-10-18  1.0.3     MRD      SB3 Off Policy Wrapper DQN, DDPG, SAC
## -- 2021-11-15  1.0.4     DA       Refactoring
## -- 2021-12-03  1.0.5     DA       Refactoring
## -- 2021-12-07  1.0.6     DA       Refactoring
## -- 2022-02-25  1.0.7     SY       Refactoring due to auto generated ID in class Dimension
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.7 (2022-02-25)

This module shows how to train with SB3 Wrapper for On-Policy Algorithm
"""

import gym
from stable_baselines3 import A2C, PPO, DQN, DDPG, SAC
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro
from mlpro.wrappers.sb3 import WrPolicySB32MLPro
from collections import deque
from pathlib import Path


# 1 Implement your own RL scenario
class MyScenario(RLScenario):
    C_NAME = 'Matrix'

    def _setup(self, p_mode, p_ada, p_logging):
        # 1 Setup environment
        # self._env   = RobotHTM(p_logging=False)
        gym_env = gym.make('CartPole-v1')
        self._env = WrEnvGYM2MLPro(gym_env, p_logging=p_logging)

        # 2 Instantiate Policy From SB3
        # env is set to None, it will be set up later inside the wrapper
        # _init_setup_model is set to False, the _setup_model() will be called inside
        # the wrapper manually

        # A2C
        # policy_sb3 = A2C(
        #             policy="MlpPolicy", 
        #             env=None,
        #             use_rms_prop=False, 
        #             _init_setup_model=False,
        #             device="cpu")

        # PPO
        policy_sb3 = PPO(
            policy="MlpPolicy",
            n_steps=5,
            env=None,
            _init_setup_model=False,
            device="cpu")

        # DQN Discrete only
        # policy_sb3 = DQN(
        #             policy="MlpPolicy", 
        #             env=None,
        #             _init_setup_model=False,
        #             device="cpu")

        # DDPG Continuous only
        # policy_sb3 = DDPG(
        #             policy="MlpPolicy", 
        #             env=None,
        #             _init_setup_model=False,
        #             device="cpu")

        # SAC Continuous only
        # policy_sb3 = SAC(
        #             policy="MlpPolicy", 
        #             env=None,
        #             _init_setup_model=False,
        #             device="cpu")

        # 3 Wrap the policy
        policy_wrapped = WrPolicySB32MLPro(
            p_sb3_policy=policy_sb3,
            p_cycle_limit=self._cycle_limit,
            p_observation_space=self._env.get_state_space(),
            p_action_space=self._env.get_action_space(),
            p_ada=p_ada,
            p_logging=p_logging)

        # 4 Setup standard single-agent with own policy
        return Agent(
            p_policy=policy_wrapped,
            p_envmodel=None,
            p_name='Smith',
            p_ada=p_ada,
            p_logging=p_logging
        )


# 2 Create scenario and start training

if __name__ == "__main__":
    # 2.1 Parameters for demo mode
    logging = Log.C_LOG_ALL
    visualize = True
    path = str(Path.home())

else:
    # 2.2 Parameters for internal unit test
    logging = Log.C_LOG_NOTHING
    visualize = False
    path = None

# 2.3 Create and run training object
training = RLTraining(
    p_scenario_cls=MyScenario,
    p_cycle_limit=1000,
    p_max_adaptations=0,
    p_max_stagnations=0,
    p_path=path,
    p_visualize=visualize,
    p_logging=logging)

training.run()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_013_model_based_reinforcement_learning.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-17  0.0.0     MRD       Creation
## -- 2021-12-17  1.0.0     MRD       Released first version
## -- 2022-01-01  1.0.1     MRD       Refactoring due to new model implementation
## -- 2022-05-20  1.0.2     MRD       Add HTMEnvModel
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.2 (2022-05-20)

This module demonstrates model-based reinforcement learning (MBRL).
"""


import torch

from mlpro.bf.ml import *
from mlpro.rl.models import *
from mlpro.rl.pool.envs.robotinhtm import RobotHTM
from stable_baselines3 import PPO
from mlpro.wrappers.sb3 import WrPolicySB32MLPro
from mlpro.rl.pool.envmodels.mlp_robotinhtm import MLPEnvModel
from mlpro.rl.pool.envmodels.htm_robotinhtm import HTMEnvModel

from pathlib import Path



class ActualTraining(RLTraining):
    C_NAME = "Actual"



# Implement RL Scenario for the actual environment to train the environment model
class ScenarioRobotHTMActual(RLScenario):
    C_NAME = "Matrix1"

    def _setup(self, p_mode, p_ada, p_logging):
        # 1 Setup environment
        self._env = RobotHTM(p_logging=True)

        policy_kwargs = dict(activation_fn=torch.nn.Tanh,
                             net_arch=[dict(pi=[128, 128], vf=[128, 128])])

        policy_sb3 = PPO(
            policy="MlpPolicy",
            n_steps=100,
            env=None,
            _init_setup_model=False,
            policy_kwargs=policy_kwargs,
            device="cpu"
        )

        policy_wrapped = WrPolicySB32MLPro(
            p_sb3_policy=policy_sb3,
            p_cycle_limit=self._cycle_limit,
            p_observation_space=self._env.get_state_space(),
            p_action_space=self._env.get_action_space(),
            p_ada=p_ada,
            p_logging=p_logging,
        )

        mb_training_param = dict(p_cycle_limit=100,
                                 p_cycles_per_epi_limit=100,
                                 p_max_stagnations=0,
                                 p_collect_states=False,
                                 p_collect_actions=False,
                                 p_collect_rewards=False,
                                 p_collect_training=False)

        # 2 Setup standard single-agent with own policy
        return Agent(
            p_policy=policy_wrapped,
            p_envmodel=HTMEnvModel(),
            p_em_mat_thsld=0.5,
            p_name="Smith1",
            p_ada=p_ada,
            p_logging=p_logging,
            **mb_training_param
        )


# 3 Train agent in scenario
if __name__ == "__main__":
    # 3.1 Parameters for demo mode
    cycle_limit = 300000
    logging     = Log.C_LOG_ALL
    visualize   = True
    path        = str(Path.home())
    plotting    = True
else:
    # 3.2 Parameters for internal unit test
    cycle_limit = 100
    logging     = Log.C_LOG_NOTHING
    visualize   = False
    path        = None
    plotting    = False

training = ActualTraining(
    p_scenario_cls=ScenarioRobotHTMActual,
    p_cycle_limit=cycle_limit,
    p_cycles_per_epi_limit=100,
    p_collect_states=True,
    p_collect_actions=True,
    p_collect_rewards=True,
    p_collect_training=True,
    p_path=path,
    p_logging=logging,
)

training.run()


# 4 Create Plotting Class
class MyDataPlotting(DataPlotting):
    def get_plots(self):
        """
        A function to plot data
        """
        for name in self.data.names:
            maxval = 0
            minval = 0
            if self.printing[name][0]:
                fig = plt.figure(figsize=(7, 7))
                raw = []
                label = []
                ax = fig.subplots(1, 1)
                ax.set_title(name)
                ax.grid(True, which="both", axis="both")
                for fr_id in self.data.frame_id[name]:
                    raw.append(np.sum(self.data.get_values(name, fr_id)))
                    if self.printing[name][1] == -1:
                        maxval = max(raw)
                        minval = min(raw)
                    else:
                        maxval = self.printing[name][2]
                        minval = self.printing[name][1]

                    label.append("%s" % fr_id)
                ax.plot(raw)
                ax.set_ylim(minval - (abs(minval) * 0.1), maxval + (abs(maxval) * 0.1))
                ax.set_xlabel("Episode")
                ax.legend(label, bbox_to_anchor=(1, 0.5), loc="center left")
                self.plots[0].append(name)
                self.plots[1].append(ax)
                if self.showing:
                    plt.show()
                else:
                    plt.close(fig)


# 5 Plotting 1 MLpro
data_printing = {
    "Cycle": [False],
    "Day": [False],
    "Second": [False],
    "Microsecond": [False],
    "Smith1": [True, -1],
}

mem = training.get_results().ds_rewards
mem_plot = MyDataPlotting(mem, p_showing=plotting, p_printing=data_printing)
mem_plot.get_plots()



## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro.pool.sarsbuffer
## -- Module  : PrioritizedBuffer
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-22  0.0.0     WB       Creation
## -- 2021-09-22  1.0.0     WB       Added PrioritizedBuffer Class and PrioritizedBufferElement,
## --                                including the required SegmentTree data structure
## -- 2021-09-26  1.0.1     WB       Bug Fix 
## -------------------------------------------------------------------------------------------------
## -- Reference
## -- https://github.com/openai/baselines/blob/master/baselines/deepq/replay_buffer.py


"""
Ver. 1.0.1 (2021-09-26)

This module provides the Prioritized Buffer based on the reference.
"""

import numpy as np
from typing import List, Callable
import random
import operator
from mlpro.rl.models import *




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class PrioritizedBufferElement(SARSElement):
    """
    Element of a State-Action-Reward-Buffer.
    """
    pass


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class PrioritizedBuffer(SARSBuffer):
    """
    Prioritized Sampling State-Action-Reward-Buffer in dictionary.
    """
    
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, p_size=1, alpha: float=0.3, beta: float=1):
    
        """
        Parameters:
            p_size (int, optional): Buffer size. Defaults to 1.
            alpha (float, optional): Prioritization level. Defaults to 0.3
            beta (float, optional): Prioritization Control. Defaults to 1. Should be increased gradualy to 1 by the end of training.
        """
        assert alpha >= 0
        assert beta >= 0
        super().__init__(p_size=p_size)
        
        self.alpha = alpha
        self.beta = beta
        
        tree_capacity = 1
        while tree_capacity < self._size:
            tree_capacity *= 2
        
        self.sum_tree = SumSegmentTree(tree_capacity)
        self.min_tree = MinSegmentTree(tree_capacity)
        self.max_priority = 1.0
    
    
## -------------------------------------------------------------------------------------------------    
    def add_element(self, p_elem:PrioritizedBufferElement):
        """
        Add element to the buffer.

        Parameters:
            p_elem (BufferElement): Element of Buffer
        """
        super().add_element(p_elem)
        idx = len(self._data_buffer)-1
        self.sum_tree[idx] = self.max_priority**self.alpha
        self.min_tree[idx] = self.max_priority**self.alpha
    
    
## -------------------------------------------------------------------------------------------------
    def _gen_sample_ind(self, p_num:int) -> list:
        """
        Generate random indices from the buffer.

        Parameters:
            p_num (int): Number of sample

        Returns:
            List of incides
        """
        buffer_length = len(self._data_buffer)
        p_sum = self.sum_tree.sum(0, buffer_length-1)
        p_list_idx = []
        segment = p_sum / buffer_length
        for i in range(p_num):
            a = segment*i
            b = segment*(i+1)
            upperbound = random.uniform(a, b)
            idx = self.sum_tree.retrieve(upperbound)
            p_list_idx.append(idx)
        return p_list_idx
        

## -------------------------------------------------------------------------------------------------
    def _extract_rows(self, p_list_idx:list):
        """
        Extract the element in the buffer based on a
        list of indices.

        Parameters:
            p_list_idx (list): List of indices

        Returns:
            Samples in dictionary
        """
        rows = {}
        for key in self._data_buffer:
            rows[key] = [self._data_buffer[key][i] for i in p_list_idx]
        p_sample = []
        buffer_length = len(self._data_buffer)
        
        p_min = self.min_tree.min()/self.sum_tree.sum()
        max_weight = (p_min*buffer_length)**(-self.beta)
        for idx in p_list_idx:
            p_sample.append(self.sum_tree[idx]/self.sum_tree.sum())
        weights = (np.array(p_sample*buffer_length)**(-self.beta))/max_weight
        
        rows['weights'] = list(weights)
        rows['p_list_idx'] = p_list_idx
        
        return rows


## -------------------------------------------------------------------------------------------------
    def get_latest(self):
        """
        Returns latest buffered element. 
        """
        try:
            return self._extract_rows([len(self._data_buffer)-1])
        except:
            return None
        
        
## -------------------------------------------------------------------------------------------------
    def get_all(self):
        """
        Return all buffered elements.
        """
        p_list_idx = [i for i in range(len(self._data_buffer))]
        return self._extract_rows(p_list_idx)


## -------------------------------------------------------------------------------------------------
    def update_priorities(self, p_list_idx:list, priorities:np.ndarray):
        """
        Updates the priority tree.
        Needs to be called during each training step, utilising the element-wise calculated loss.
        """
        assert len(p_list_idx) == len(priorities)
        assert np.min(priorities) > 0 
        assert min(p_list_idx) >= 0
        assert max(p_list_idx) <= len(self._data_buffer)
        
        new_priorities = priorities**self.alpha
        for i in range(len(p_list_idx)):
            self.sum_tree[p_list_idx[i]] = new_priorities[i]
            self.min_tree[p_list_idx[i]] = new_priorities[i]
        
        self.max_priority = max(self.max_priority, np.max(new_priorities))


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SegmentTree:
    """ 
    Reference:
    https://github.com/openai/baselines/blob/master/baselines/common/segment_tree.py
    Attributes:
        capacity (int)
        tree (list)
        operation (function)
    """


## -------------------------------------------------------------------------------------------------
    def __init__(self, capacity: int, operation: Callable, init_value: float):
        assert (
            capacity > 0 and capacity & (capacity - 1) == 0
        ), "capacity must be positive and a power of 2."
        self.capacity = capacity
        self.tree = [init_value for _ in range(2 * capacity)]
        self.operation = operation
        
        
## -------------------------------------------------------------------------------------------------
    def _operate_helper(
        self, start: int, end: int, node: int, node_start: int, node_end: int
    ) -> float:
        """Returns result of operation in segment."""
        if start == node_start and end == node_end:
            return self.tree[node]
        mid = (node_start + node_end) // 2
        if end <= mid:
            return self._operate_helper(start, end, 2 * node, node_start, mid)
        else:
            if mid + 1 <= start:
                return self._operate_helper(start, end, 2 * node + 1, mid + 1, node_end)
            else:
                return self.operation(
                    self._operate_helper(start, mid, 2 * node, node_start, mid),
                    self._operate_helper(mid + 1, end, 2 * node + 1, mid + 1, node_end),
                )


## -------------------------------------------------------------------------------------------------
    def operate(self, start: int = 0, end: int = 0) -> float:
        """Returns result of applying 'self.operation'."""
        if end <= 0:
            end += self.capacity
        end -= 1

        return self._operate_helper(start, end, 1, 0, self.capacity - 1)


## -------------------------------------------------------------------------------------------------
    def __setitem__(self, idx: int, val: float):
        """Set value in tree."""
        idx += self.capacity
        self.tree[idx] = val

        idx //= 2
        while idx >= 1:
            self.tree[idx] = self.operation(self.tree[2 * idx], self.tree[2 * idx + 1])
            idx //= 2


## -------------------------------------------------------------------------------------------------
    def __getitem__(self, idx: int) -> float:
        """Get real value in leaf node of tree."""
        assert 0 <= idx < self.capacity

        return self.tree[self.capacity + idx]
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SumSegmentTree(SegmentTree):
    """ 
    Reference:
    https://github.com/openai/baselines/blob/master/baselines/common/segment_tree.py
    """


## -------------------------------------------------------------------------------------------------
    def __init__(self, capacity: int):
        super(SumSegmentTree, self).__init__(
            capacity=capacity, operation=operator.add, init_value=0.0
        )


## -------------------------------------------------------------------------------------------------
    def sum(self, start: int = 0, end: int = 0) -> float:
        """Returns arr[start] + ... + arr[end]."""
        return super(SumSegmentTree, self).operate(start, end)


## -------------------------------------------------------------------------------------------------
    def retrieve(self, upperbound: float) -> int:
        """Find the highest index `i` about upper bound in the tree"""
        # TODO: Check assert case and fix bug
        assert 0 <= upperbound <= self.sum() + 1e-5, "upperbound: {}".format(upperbound)

        idx = 1

        while idx < self.capacity:  # while non-leaf
            left = 2 * idx
            right = left + 1
            if self.tree[left] > upperbound:
                idx = 2 * idx
            else:
                upperbound -= self.tree[left]
                idx = right
        return idx - self.capacity
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class MinSegmentTree(SegmentTree):
    """ 
    Reference:
    https://github.com/openai/baselines/blob/master/baselines/common/segment_tree.py
    """


## -------------------------------------------------------------------------------------------------
    def __init__(self, capacity: int):
        super(MinSegmentTree, self).__init__(
            capacity=capacity, operation=min, init_value=float("inf")
        )


## -------------------------------------------------------------------------------------------------
    def min(self, start: int = 0, end: int = 0) -> float:
        """Returns min(arr[start], ...,  arr[end])."""
        return super(MinSegmentTree, self).operate(start, end)


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.wrappers
## -- Module  : hyperopt.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-07  0.0.0     SY       Creation 
## -- 2021-12-08  1.0.0     SY       Release of first version
## -- 2022-01-21  1.0.1     DA       Fixed some bugs
## -- 2022-01-27  1.0.2     SY       Wrapper enhancement
## -- 2022-01-28  1.0.3     SY       - Move save() and save_line() to HyperParamTuner
## --                                - Store the parameters and its result for each trial
## -- 2022-02-28  1.0.4     SY       Enhancement due to redefinition of _init_hyperparam() in Agent
## -- 2022-03-02  1.0.5     SY       Refactoring
## -- 2022-03-24  1.0.6     SY       Refactoring
## -- 2022-03-25  1.0.7     SY       Change methods names (SetupSpaces to setup_spaces)
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.7 (2022-03-25)
This module provides a wrapper class for hyperparameter tuning by reusing Hyperopt framework
"""


from hyperopt import *
from mlpro.bf.ml import *
from mlpro.bf.math import *
from mlpro.bf.various import *
from mlpro.rl.models import *
from mlpro.gt.models import *
import os




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrHPTHyperopt(HyperParamTuner, ScientificObject):
    """
    This class is a ready to use wrapper class for Hyperopt framework. 
    Objects of this type can be treated as a hyperparameter tuner object.
    
    Parameters
    ----------
    p_logging:
        Log level (see constants for log levels)
    p_algo : str, optional    
        Selection of a hyperparameter tuning algorithm, default: C_ALGO_RAND
    p_ids : list of str, optional
        List of hyperparameter ids to be tuned, otherwise all hyperparameters, default: None
           
    Attributes
    ----------
    C_NAME: str
        Name of the class.
    C_ALGO_TPE: str
        Refer to Tree of Parzen Estimators (TPE) algorithm.
    C_ALGO_RAND: str
        Refer to Random Grid Search algorithm.
    """
    
    C_NAME              = 'Hyperopt'
    
    C_ALGO_TPE          = 'TPE'
    C_ALGO_RAND         = 'RND'
        
    C_SCIREF_TYPE       = ScientificObject.C_SCIREF_TYPE_PROCEEDINGS
    C_SCIREF_AUTHOR     = "James Bergstra, Dan Yamins, David D. Cox"
    C_SCIREF_TITLE      = "Hyperopt: A Python Library for Otimizing the Hyperparameters of Machine Learning Algorithms"
    C_SCIREF_CONFERENCE = "Proceedings of the 12th Python in Science Conference"
    C_SCIREF_YEAR       = "2013"
    C_SCIREF_PAGES      = "13-19"
    C_SCIREF_DOI        = "10.25080/Majora-8b375195-003"
    C_SCIREF_EDITOR     = "Stefan van der Walt, Jarrod Millman, Katy Huff"
    
    C_LOG_SEPARATOR = '------------------------------------------------------------------------------'
    

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_logging=Log.C_LOG_ALL, p_algo=C_ALGO_RAND, p_ids=None):
        super().__init__(p_logging=p_logging)

        if p_algo is None:
            raise ParamError('Mandatory parameter p_algo is not supplied')
        else:
            self._algo = p_algo
        
        self._ids = p_ids
        self.num_trials = 0
        self.main_path = None
        
        self.log(self.C_LOG_TYPE_I, 'Hyperopt configuration is successful')


## -------------------------------------------------------------------------------------------------
    def _maximize(self) -> TrainingResults:
        """
        This method is a place to setup a hp tuner based on hp structure of the model
        and run the hp tuner.

        Returns
        -------
        best_result : float
            The best result after a number of evaluations.

        """
        
        if self._training_cls is None:
            raise ParamError('Mandatory parameter self._training_cls is not supplied')
        
        if self._num_trials <= 0:
            raise ParamError('Parameter self._num_trials must be greater than 0')
        
        if self._root_path is None:
            raise ParamError('Mandatory parameter self._root_path is not supplied')
        
        if self._training_param is None:
            raise ParamError('Mandatory parameter self._training_param is not supplied')
        
        # change root path in training param
        self._training_param['p_training_param']['p_path'] = self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Base (Preparation)'
        if not os.path.exists(self._training_param['p_training_param']['p_path']):
            os.mkdir(self._root_path+os.sep+'HyperparameterTuning')
            os.mkdir(self._training_param['p_training_param']['p_path'])
        
        # ignore collecting data during tuning to save tuning time and memory
        self._training_param['p_training_param']['p_collect_states'] = False
        self._training_param['p_training_param']['p_collect_actions'] = False
        self._training_param['p_training_param']['p_collect_rewards'] = False
        self._training_param['p_training_param']['p_logging'] = Log.C_LOG_NOTHING
        self._training_param['p_training_param']['p_visualize'] = False
        self._training_param['p_training_param']['p_collect_eval'] = True
        
        # instantiate a scenario class and define the model in a variable
        training_cls = self._training_cls(**self._training_param['p_training_param'])
        self._model = training_cls._scenario._model
        
        # setup spaces according to Hyperopt requirements
        spaces = self.setup_spaces()
        
        # prepare a data storing class
        self.HPDataStoring = DataStoring(self.variables)
        self.HPDataStoring.add_frame('HP_0')
        
        # algorithm selection
        if self._algo == 'TPE':
            self.algo = tpe.suggest
        elif self._algo == 'RND':
            self.algo = rand.suggest
        
        # run the trials and gain the highest score
        trials = Trials()
        best_param = fmin(self._ofct_hyperopt, spaces, self.algo, self._num_trials, trials=trials)
        best_result = trials.results[np.argmin([r['loss'] for r in trials.results])]['loss']
        self.save(best_param, -best_result, 'best_parameters.csv')
        
        return -best_result


## -------------------------------------------------------------------------------------------------
    def _ofct_hyperopt(self, p_params):
        """
        This method is a place to run the evaluations by getting next set of hps from the tuner,
        inducting hps to the model, and running the the objective function.

        Returns
        -------
        result : float
            The result of an evaluations.
            
        """
        
        self.log(self.C_LOG_TYPE_I, 'Trial number '+str(self.num_trials)+' has started')
        self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')
        
        # change root path in training param
        self._training_param['p_training_param']['p_path'] =  self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Trial_'+str(self.num_trials)
        if not os.path.exists(self._training_param['p_training_param']['p_path']):
            os.mkdir(self._training_param['p_training_param']['p_path'])
        
        # instantiate a scenario class
        training_cls = self._training_cls(**self._training_param['p_training_param'])
        self._model = training_cls._scenario._model
        
        # change the parameter according to p_params generated by hyperopt
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            self._model._hyperparam_tuple.set_value(_id, p_params[x])
            
        # run the scenario and retrieve the high score
        result = training_cls.run()
        
        # store trial parameters and the trial's result
        self.HPDataStoring.memorize('Trial', 'HP_0', self.num_trials)
        self.HPDataStoring.memorize('Highscore', 'HP_0', result.highscore)
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            hp_name_short = self._model._hyperparam_tuple.get_related_set().get_dim(_id).get_name_short()
            self.HPDataStoring.memorize(hp_name_short+'_'+str(x), 'HP_0', p_params[x])
            
        self.num_trials += 1
        
        self.log(self.C_LOG_TYPE_I, 'Trial number '+str(self.num_trials)+' has finished')
        self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')
        
        return -(result.highscore)


## -------------------------------------------------------------------------------------------------
    def setup_spaces(self):
        """
        This method is used to setup the hyperparameter spaces, including the tuning boundaries and a suitable discrete value.
        The hyperparameter should be bounded both above and below.
        We are using the "quantized" continuous distributions for natural and integer numbers.
        Meanwhile the real numbers are not quantized.
        For different parameter expressions, please redefined this method and check http://hyperopt.github.io/hyperopt/getting-started/search_spaces/!
        For big data handling, please redifined this method!
        
        Returns
        -------
        spaces : list
            List of parameter expressions.

        """
        
        if self._model._hyperparam_tuple is None:
            self._model._init_hyperparam()
        
        spaces = []
        for x, _id in enumerate(self._model._hyperparam_tuple.get_dim_ids()):
            hp_object = self._model._hyperparam_tuple.get_related_set().get_dim(_id)
            hp_boundaries = hp_object.get_boundaries()
            if not hp_boundaries:
                raise ImplementationError('Missing boundary of a hyperparameter!')
            else:
                hp_low = hp_boundaries[0]
                hp_high = hp_boundaries[1]
                if hp_object._base_set == Dimension.C_BASE_SET_N or hp_object._base_set == Dimension.C_BASE_SET_Z:
                    spaces.append(hp.quniform(hp_object.get_name_short()+'_'+str(x),hp_low,hp_high,1))
                elif hp_object._base_set == Dimension.C_BASE_SET_R:
                    spaces.append(hp.uniform(hp_object.get_name_short()+'_'+str(x),hp_low,hp_high))
                else:
                    raise ImplementationError('Missing a short name of a hyperparameter!')
            self.variables.append(hp_object.get_name_short()+'_'+str(x))
        
        self.log(self.C_LOG_TYPE_I, 'Spaces for hyperopt is ready')
        self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')
        
        return spaces



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl
## -- Module  : models_agents.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-18  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Release of first version
## -- 2021-06-13  1.0.1     DA       New method Agent.reset_episode() and call in class Training;
## --                                Some bugfixes in class MultiAgent
## -- 2021-06-25  1.0.2     DA       New methods Agent.get_name(), Agent.set_name()
## -- 2021-07-01  1.0.3     DA       Bugfixes in classes MultiAgent
## -- 2021-08-26  1.1.0     DA       New class Policy; enhancements of Agent class;
## --                                Incompatible changes on classes Agent, MultiAgent
## -- 2021-08-28  1.1.1     DA       Bugfixes and minor improvements
## -- 2021-09-11  1.1.2     MRD      Change Header information to match our new library name
## -- 2021-09-18  1.2.0     MRD      Implement new SARBuffer class and SARBufferelement. The buffer
## --                                is now working in dictionary. Now the SARBuffer is inside
## --                                the Policy and EnvModel instead of the Agent.
## --                                Added "Done" as default input for Agent.adapt()
## -- 2021-10-05  1.2.1     DA       Various changes:
## --                                - New class ActionPlanner
## --                                - Class Agent: preparation for model-based mode
## -- 2021-10-05  1.2.2     SY       Bugfixes and minor improvements
## -- 2021-10-18  1.2.3     DA       Refactoring Policy/Agent/MultiAgent: state space renamed to 
## --                                observation space
## -- 2021-11-14  1.3.0     DA       Model-based Agent functionality 
## -- 2021-11-26  1.3.1     DA       Minor changes
## -- 2021-12-17  1.3.2     DA       Added method MultiAgent.get_agent()
## -- 2021-12-30  1.4.0     DA       - Class Agent: added internal model-based policy training
## --                                - Class ActionPlanner completed
## --                                - Standardization of all docstrings
## -- 2022-01-01  1.4.1     MRD      Refactoring and Fixing some bugs
## -- 2022-01-28  1.4.2     SY       - Added switch_adaptivity method in MultiAgent class
## --                                - Update _adapt method in MultiAgent class
## -- 2022-02-17  1.5.0     DA/SY    Class Agent: redefinition of method _init_hyperparam()
## -- 2022-02-24  1.5.1     SY       Class MultiAgent: redefinition of method _init_hyperparam()
## -- 2022-02-27  1.5.2     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-03-02  1.5.3     SY       Class MultiAgent: remove init_hyperparam(), update add_agent()
## -- 2022-03-02  1.5.4     DA       Reformatting
## -- 2022-03-07  1.5.5     SY       Minor Improvement on Class MultiAgent
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.5.5 (2022-03-07) 

This module provides model classes for policies, model-free and model-based agents and multi-agents.
"""

from mlpro.rl.models_env import *
from mlpro.rl.models_train import RLScenario, RLTraining





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Policy(Model):
    """
    This class represents the policy of a single-agent. It is adaptive and can be trained with
    State-Action-Reward (SAR) data that will be expected as a SAR buffer object. 
    The three main learning paradigms of machine learning to train a policy are supported:

    a) Training by Supervised Learning
    The entire SAR data set inside the SAR buffer shall be adapted.

    b) Training by Reinforcement Learning
    The latest SAR data record inside the SAR buffer shall be adapted.

    c) Training by Unsupervised Learning
    All state data inside the SAR buffer shall be adapted.

    Furthermore, a policy class can compute actions from states.

    Hyperparameters of the policy should be stored in the internal object self._hp_list, so that
    they can be tuned from outside. Optionally a policy-specific callback method can be called on 
    changes. For more information see class HyperParameterList.

    Parameters
    ----------
    p_observation_space : MSpace     
        Subspace of an environment that is observed by the policy.
    p_action_space : MSpace
        Action space object.
    p_buffer_size : int           
        Size of internal buffer. Default = 1.
    p_ada : bool               
        Boolean switch for adaptivity. Default = True.
    p_logging
        Log level (see constants of class Log). Default = Log.C_LOG_ALL.

    """

    C_TYPE = 'Policy'
    C_NAME = '????'
    C_BUFFER_CLS = SARSBuffer

## -------------------------------------------------------------------------------------------------
    def __init__(self,
                 p_observation_space: MSpace,
                 p_action_space: MSpace,
                 p_buffer_size=1,
                 p_ada=True,
                 p_logging=Log.C_LOG_ALL):
        super().__init__(p_buffer_size=p_buffer_size, p_ada=p_ada, p_logging=p_logging)
        self._observation_space = p_observation_space
        self._action_space = p_action_space
        self.set_id(0)


## -------------------------------------------------------------------------------------------------
    def get_observation_space(self) -> MSpace:
        return self._observation_space


## -------------------------------------------------------------------------------------------------
    def get_action_space(self) -> MSpace:
        return self._action_space


## -------------------------------------------------------------------------------------------------
    def get_id(self):
        return self._id


## -------------------------------------------------------------------------------------------------
    def set_id(self, p_id):
        self._id = p_id


## -------------------------------------------------------------------------------------------------
    def compute_action(self, p_obs: State) -> Action:
        """
        Specific action computation method to be redefined. 

        Parameters
        ----------
        p_obs : State
            Observation data.

        Returns
        -------
        action : Action
            Action object.

        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def _adapt(self, *p_args) -> bool:
        """
        Adapts the policy based on State-Action-Reward-State (SARS) data.

        Parameters
        ----------
        p_arg[0] : SARSElement
            Object of type SARSElement.

        Returns
        -------
        adapted : bool
            True, if something has been adapted. False otherwise.

        """

        raise NotImplementedError





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class ActionPlanner(Log):
    """
    Template class for action planning algorithms to be used as part of model-based planning agents. 
    The goal is to find the shortest sequence of actions that leads to a maximum reward.

    Parameters
    ----------
    p_state_thsld : float
        Threshold for metric difference between two states to be equal. Default = 0.00000001.
    p_logging
        Log level (see constants of class Log). Default = Log.C_LOG_ALL.

    """

    C_TYPE = 'Action Planner'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_state_thsld=0.00000001, p_logging=Log.C_LOG_ALL):
        super().__init__(p_logging=p_logging)
        self._depth_limit = 0
        self._width_limit = 0
        self._policy = None
        self._env_model = None
        self._action_path = None
        self._state_thsld = p_state_thsld


## -------------------------------------------------------------------------------------------------
    def setup(self,
              p_policy: Policy,
              p_envmodel: EnvModel,
              p_depth_limit=0,
              p_width_limit=0):
        """
        Setup of action planner object in concrete planning scenario. Must be called before first
        planning. Optional custom method _setup() is called at the end.

        Parameters
        ----------
        p_policy : Policy
            Policy of an agent.
        p_envmodel : EnvModel
            Environment model.
        p_depth_limit : int             
            Optional static maximum planning depth (=length of action path to be predicted). Can
            be overridden by method compute_action(). Default=0. 
        p_width_limit : int
            Optional static maximum planning width (=number of alternative actions per planning level).
            Can be overridden by method compute_action(). Default=0. 

        """

        self._policy = p_policy
        self._envmodel = p_envmodel
        self._depth_limit = p_depth_limit
        self._width_limit = p_width_limit
        self._path_id = 0
        self.clear_action_path()
        self._action_path = None
        self._setup()


## -------------------------------------------------------------------------------------------------
    def _setup(self):
        """
        Optional custom setup method.
        """

        pass


## -------------------------------------------------------------------------------------------------
    def compute_action(self,
                       p_obs: State,
                       p_depth_limit=0,
                       p_width_limit=0) -> Action:
        """
        Computes a path of actions with defined length that maximizes the reward of the given 
        environment model. The planning algorithm itself is to be implemented in the custom method
        _plan_action().
        
        Parameters
        ----------
        p_obs : State
            Observation data.
        p_depth_limit : int             
            Optional dynamic maximum planning depth (=length of action path to be predicted) that 
            overrides the static limit of method setup(). Default=0 (no override).
        p_width_limit : int
            Optional dynamic maximum planning width (=number of alternative actions per planning level)
            that  overrides the static limit of method setup(). Default=0 (no override).

        Returns
        -------
        action : Action
            Best action as result of the planning process.

        """

        if (self._policy is None) or (self._envmodel is None):
            raise RuntimeError('Please call method setup() first')

        if (p_depth_limit > 0) and (p_depth_limit != self._depth_limit):
            self._depth_limit = p_depth_limit
            self._action_path = None

        if p_width_limit > 0:
            self._width_limit = p_width_limit

        if (self._depth_limit <= 0) or (self._width_limit <= 0):
            raise RuntimeError('Please set planning depth and width')

        # Check: Re-planning required?
        replan = self._action_path is None
        replan = replan or (self._path_id >= len(self._action_path))

        if not replan:
            # Check: Is the next action of action path suitable?
            path_data = self._action_path.get_all()
            obs_buffered = path_data['state'][self._path_id]
            replan = self._policy.get_observation_space().distance(p_obs, obs_buffered) > self._state_thsld

        if replan:
            # (Re-)Planning of action path
            self._path_id = 0
            self._action_path.clear()
            self._action_path = self._plan_action(p_obs)
            if (self._action_path is None) or (len(self._action_path) == 0):
                # Planning returned nothing -> direct action computation as fallback solution
                return self._policy.compute_action(p_obs)

        # Next action of action path can be used
        action = path_data['action'][self._path_id]
        self._path_id += 1
        return action


## -------------------------------------------------------------------------------------------------
    def _plan_action(self, p_obs: State) -> SARSBuffer:
        """
        Custom planning algorithm to fill the internal action path (self._action_path). Search width
        and depth are restricted by the attributes self._width_limit and self._depth_limit.

        Parameters
        ----------
        p_obs : State
            Observation data.

        Returns
        -------
        action_path : SARSBuffer
            Sequence of SARSElement objects with included actions that lead to the best possible reward.

        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
    def clear_action_path(self):
        if self._action_path is not None:
            self._action_path.clear()





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLScenarioMBInt(RLScenario):
    """
    Internal use in class Agent. Intended for the training of the policy with the environment model of
    a model-based (single) agent.

    """

    C_NAME = 'MB(intern)'

## -------------------------------------------------------------------------------------------------
    def _setup(self, **p_kwargs) -> Model:
        # Pseudo-implementation
        self._env = EnvBase(p_logging=Log.C_LOG_NOTHING)
        return Model(p_logging=Log.C_LOG_NOTHING)


## -------------------------------------------------------------------------------------------------
    def setup_ext(self, p_env: EnvBase, p_policy: Policy):
        self._model = Agent(p_policy=p_policy)
        self._agent = self._model
        self._env = p_env





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Agent(Policy):
    """
    This class represents a single agent model.

    Parameters
    ----------
    p_policy : Policy
        Policy object.
    p_envmodel : EnvModel
        Optional environment model object. Default = None.
    p_em_mat_thsld : float
        Optional threshold for environment model maturity (whether the envmodel is 'good'
        enough to be used to train the policy). Default = 0.9.
    p_action_planner : ActionPlanner   
        Optional action planner object (obligatory for model based agents). Default = None.
    p_planning_depth : int    
        Optional planning depth (obligatory for model based agents). Default = 0.
    p_planning_width : int   
        Optional planning width (obligatory for model based agents). Default = 0.
    p_name : str             
        Optional name of agent. Default = ''.
    p_id : int               
        Optional unique agent id (especially important for multi-agent scenarios). Default = 0.
    p_ada : bool               
        Boolean switch for adaptivity. Default = True.
    p_logging          
        Log level (see constants of class mlpro.bf.various.Log). Default = Log.C_LOG_ALL.
    p_mb_training_param : dict
        Optional parameters for internal policy training with environment model (see parameters of
        class RLTraining). Hyperparameter tuning and data logging is not supported here. The suitable
        scenario class is added internally.

    """

    C_TYPE = 'Agent'
    C_NAME = ''

## -------------------------------------------------------------------------------------------------
    def __init__(self,
                 p_policy: Policy,
                 p_envmodel: EnvModel = None,
                 p_em_mat_thsld=0.9,
                 p_action_planner: ActionPlanner = None,
                 p_planning_depth=0,
                 p_planning_width=0,
                 p_name='',
                 p_id=0,
                 p_ada=True,
                 p_logging=Log.C_LOG_ALL,
                 **p_mb_training_param):

        if p_name != '':
            self.set_name(p_name)
        else:
            self.set_name(self.C_NAME)

        if p_envmodel is not None:
            if len(p_mb_training_param) == 0:
                raise ParamError('Please provide parameters for model-based training in parameter p_mb_training_param')

            self._mb_training_param = p_mb_training_param.copy()
            self._mb_training_param['p_scenario_cls'] = RLScenarioMBInt
            self._mb_training_param['p_visualize'] = False
            self._mb_training_param['p_logging'] = p_logging
            if 'p_collect_states' not in self._mb_training_param:
                self._mb_training_param['p_collect_states'] = False
            if 'p_collect_actions' not in self._mb_training_param:
                self._mb_training_param['p_collect_actions'] = False
            if 'p_collect_rewards' not in self._mb_training_param:
                self._mb_training_param['p_collect_rewards'] = False
            if 'p_collect_eval' not in self._mb_training_param:
                self._mb_training_param['p_collect_eval'] = False

            # Hyperparameter tuning is disabled here
            if 'p_hpt' in self._mb_training_param:
                self._mb_training_param.pop('p_hpt')
            if 'p_hpt_trials' in self._mb_training_param:
                self._mb_training_param.pop('p_hpt_trials')

        if (p_action_planner is not None) and (p_envmodel is None):
            raise ParamError('Agents using an action planner also need an environment model')

        self._previous_observation = None
        self._previous_action = None
        self._policy = p_policy
        self._action_space = self._policy.get_action_space()
        self._observation_space = self._policy.get_observation_space()
        self._envmodel = p_envmodel
        self._em_mat_thsld = p_em_mat_thsld
        self._action_planner = p_action_planner
        self._planning_depth = p_planning_depth
        self._planning_width = p_planning_width

        self._set_id(p_id)

        Log.__init__(self, p_logging)
        self.switch_logging(p_logging)
        self.switch_adaptivity(p_ada)

        self.clear_buffer()

        if self._action_planner is not None:
            self._action_planner.setup(p_policy=self._policy,
                                       p_envmodel=self._envmodel,
                                       p_depth_limit=self._planning_depth,
                                       p_width_limit=self._planning_width)


## -------------------------------------------------------------------------------------------------
    def _set_id(self, p_id):
        super().set_id(p_id)
        self._policy.set_id(p_id)


## -------------------------------------------------------------------------------------------------
    def set_id(self, p_id):
        """
        The unique agent id will be defined while instantiation and can't be changed anymore.
        """

        pass


## -------------------------------------------------------------------------------------------------
    def get_name(self):
        return self._name


## -------------------------------------------------------------------------------------------------
    def set_name(self, p_name):
        self._name = p_name
        self.C_NAME = p_name


## -------------------------------------------------------------------------------------------------
    def _init_hyperparam(self, **p_par):
        self._hyperparam_space = self._policy.get_hyperparam().get_related_set().copy(p_new_dim_ids=False)
        if self._envmodel is not None:
            self._hyperparam_space.append(self._envmodel.get_hyperparam().get_related_set(), p_new_dim_ids=False)

        self._hyperparam_tuple = HyperParamDispatcher(p_set=self._hyperparam_space)
        self._hyperparam_tuple.add_hp_tuple(self._policy.get_hyperparam())
        if self._envmodel is not None:
            self._hyperparam_tuple.add_hp_tuple(self._envmodel.get_hyperparam())

        
## -------------------------------------------------------------------------------------------------
    def switch_logging(self, p_logging):
        super().switch_logging(p_logging)
        self._policy.switch_logging(p_logging)


## -------------------------------------------------------------------------------------------------
    def switch_adaptivity(self, p_ada: bool):
        super().switch_adaptivity(p_ada)
        try:
            self._policy.switch_adaptivity(p_ada)
        except AttributeError:
            pass


## -------------------------------------------------------------------------------------------------
    def set_log_level(self, p_level):
        super().set_log_level(p_level)
        self._policy.set_log_level(p_level)


## -------------------------------------------------------------------------------------------------
    def get_observation_space(self) -> MSpace:
        return self._policy.get_observation_space()


## -------------------------------------------------------------------------------------------------
    def get_action_space(self) -> MSpace:
        return self._policy.get_action_space()


## -------------------------------------------------------------------------------------------------
    def _extract_observation(self, p_state: State) -> State:
        if p_state.get_related_set() == self.get_observation_space():
            return p_state

        obs_space = self.get_observation_space()
        obs_dim_ids = obs_space.get_dim_ids()
        observation = State(obs_space)

        for dim_id in obs_dim_ids:
            p_state_ids = p_state.get_dim_ids()
            obs_idx = obs_space.get_dim_ids().index(dim_id)
            observation.set_value(dim_id, p_state.get_value(p_state_ids[obs_idx]))

        return observation


## -------------------------------------------------------------------------------------------------
    def set_random_seed(self, p_seed=None):
        self._policy.set_random_seed(p_seed)
        if self._envmodel is not None:
            self._envmodel.set_random_seed(p_seed)


## -------------------------------------------------------------------------------------------------
    def compute_action(self, p_state: State) -> Action:
        """
        Default implementation of a single agent.

        Parameters
        ----------
        p_state : State        
            State of the related environment.

        Returns
        -------
        action : Action
            Action object.

        """

        # 0 Intro
        self.log(self.C_LOG_TYPE_I, 'Action computation started')
        observation = self._extract_observation(p_state)

        # 1 Action computation
        if self._action_planner is None:
            # 1.1 W/o action planner
            action = self._policy.compute_action(observation)

        else:
            # 1.2 With action planner
            action = self._action_planner.compute_action(observation)

        # 2 Outro
        self.log(self.C_LOG_TYPE_I, 'Action computation finished')
        self._previous_observation = observation
        self._previous_action = action
        return action


## -------------------------------------------------------------------------------------------------
    def _adapt(self, *p_args) -> bool:
        """
        Default adaptation implementation of a single agent.

        Parameters
        ----------
        p_args[0] : State       
            State object.
        p_args[1] : Reward     
            Reward object.
 
        Returns
        -------
        result : bool
            True, if something has been adapted. False otherwise.

        """

        # 1 Check: Adaptation possible?
        if self._previous_observation is None:
            self.log(self.C_LOG_TYPE_I, 'Adaption: previous observation is None -> adaptivity skipped')
            return False

        # 2 Extract agent specific observation data from state
        state = p_args[0]
        reward = p_args[1]
        observation = self._extract_observation(state)
        adapted = False

        # 3 Adaptation
        if self._envmodel is None:
            # 3.1 Model-free adaptation
            adapted = self._policy.adapt(
                SARSElement(self._previous_observation, self._previous_action, reward, observation))

        else:
            # 3.2 Model-based adaptation
            adapted = self._envmodel.adapt(
                SARSElement(self._previous_observation, self._previous_action, reward, observation))

            if self._envmodel.get_maturity() >= self._em_mat_thsld:
                adapted = adapted or self._adapt_policy_by_model()

        return adapted


## -------------------------------------------------------------------------------------------------
    def _adapt_policy_by_model(self):
        self.log(self.C_LOG_TYPE_I, 'Model-based policy training')
        training = RLTraining(**self._mb_training_param)
        training.get_scenario().setup_ext(p_env=self._envmodel, p_policy=self._policy)

        # The RLTraining need to be adjusted again due to setup_ext()
        # And also due to model_train.py line 595 only executed on RLTraining init
        # Not after the setup_ext
        training._env = training.get_scenario().get_env()
        training._agent = training.get_scenario().get_agent()

        return training.run().num_adaptations > 0


## -------------------------------------------------------------------------------------------------
    def clear_buffer(self):
        self._policy.clear_buffer()
        if self._envmodel is not None:
            self._envmodel.clear_buffer()





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class MultiAgent(Agent):
    """
    Multi-Agent.

    Parameters
    ----------
    p_name : str
        Name of agent. Default = ''.
    p_ada : bool               
        Boolean switch for adaptivity. Default = True.
    p_logging
        Log level (see constants of class Log). Default = Log.C_LOG_ALL.

    """

    C_TYPE = 'Multi-Agent'
    C_NAME = ''
    C_SUFFIX = '.cfg'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_name='', p_ada=True, p_logging=True):
        self._agents = []
        self._agent_ids = []
        self.set_name(p_name)

        Log.__init__(self, p_logging)
        self.switch_logging(p_logging)
        self.switch_adaptivity(p_ada)
        self._set_adapted(False)


## -------------------------------------------------------------------------------------------------
    def switch_logging(self, p_logging) -> None:
        Log.switch_logging(self, p_logging=p_logging)

        for agent_entry in self._agents:
            agent_entry[0].switch_logging(p_logging)


## -------------------------------------------------------------------------------------------------
    def switch_adaptivity(self, p_ada: bool):
        super().switch_adaptivity(p_ada)
        
        for agent_entry in self._agents:
            agent_entry[0].switch_adaptivity(p_ada)

    
## -------------------------------------------------------------------------------------------------
    def set_log_level(self, p_level):
        Log.set_log_level(self, p_level)

        for agent_entry in self._agents:
            agent_entry[0].set_log_level(p_level)


## -------------------------------------------------------------------------------------------------
    def get_filename(self) -> str:
        return self.C_TYPE + ' ' + self.C_NAME + self.C_SUFFIX


## -------------------------------------------------------------------------------------------------
    def load(self, p_path, p_filename=None) -> bool:
        # load all subagents
        for i, agent_entry in enumerate(self._agents):
            agent = agent_entry[0]
            agent_name = agent.C_TYPE + ' ' + agent.C_NAME + '(' + str(i) + ')'

            if agent.load(p_path, agent_name + agent.C_SUFFIX):
                self.log(Log.C_LOG_TYPE_I, agent_name + ' loaded')
            else:
                self.log(Log.C_LOG_TYPE_E, agent_name + ' not loaded')
                return False

        return True


## -------------------------------------------------------------------------------------------------
    def save(self, p_path, p_filename=None) -> bool:
        # save all subagents
        for i, agent_entry in enumerate(self._agents):
            agent = agent_entry[0]
            agent_name = agent.C_TYPE + ' ' + agent.C_NAME + '(' + str(i) + ')'

            if agent.save(p_path, agent_name + agent.C_SUFFIX):
                self.log(Log.C_LOG_TYPE_I, agent_name + ' saved')
            else:
                self.log(Log.C_LOG_TYPE_E, agent_name + ' not saved')
                return False

        return True


## -------------------------------------------------------------------------------------------------
    def add_agent(self, p_agent: Agent, p_weight=1.0) -> None:
        """
        Adds agent object to internal list of agents. 

        Parameters
        ----------
        p_agent : Agent
            Agent object to be added.
        p_weight : float         
            Optional weight for the agent. Default = 1.0.

        """

        p_agent.switch_adaptivity(self._adaptivity)
        self._agent_ids.append(p_agent.get_id())
        self._agents.append([p_agent, p_weight])
        p_agent.set_name(str(p_agent.get_id()) + ' ' + p_agent.get_name())
        self.log(Log.C_LOG_TYPE_I, p_agent.C_TYPE + ' ' + p_agent.get_name() + ' added.')

        agent_model = self._agents[p_agent.get_id()][0]
        
        if agent_model._policy.get_hyperparam() is not None:
            if p_agent.get_id() == 0:
                self._hyperparam_space = agent_model._policy.get_hyperparam().get_related_set().copy(p_new_dim_ids=False)
            else:
                self._hyperparam_space.append(agent_model._policy.get_hyperparam().get_related_set(), p_new_dim_ids=False)
        
        if agent_model._envmodel is not None:
            try:
                self._hyperparam_space.append(agent_model._envmodel.get_hyperparam().get_related_set())
            except:
                pass
 
        if '_hyperparam_space' in dir(self):
            self._hyperparam_tuple = HyperParamDispatcher(p_set=self._hyperparam_space)
            
            for x, mod in enumerate(self._agents):
                self._hyperparam_tuple.add_hp_tuple(mod[0]._policy.get_hyperparam())
                
                if mod[0]._envmodel is not None:
                    try:
                        self._hyperparam_tuple.add_hp_tuple(mod[0]._envmodel.get_hyperparam())
                    except:
                          pass
        

## -------------------------------------------------------------------------------------------------
    def get_agents(self):
        return self._agents


## -------------------------------------------------------------------------------------------------
    def get_agent(self, p_agent_id):
        """
        Returns information of a single agent.

        Returns
        -------
        agent_info : list
            agent_info[0] is the agent object itself and agent_info[1] it's weight
            
        """

        return self._agents[self._agent_ids.index(p_agent_id)]

    
## -------------------------------------------------------------------------------------------------
    def get_observation_space(self) -> MSpace:
        return None

    
## -------------------------------------------------------------------------------------------------
    def get_action_space(self) -> MSpace:
        return None

    
## -------------------------------------------------------------------------------------------------
    def set_random_seed(self, p_seed=None):
        for i, agent_entry in enumerate(self._agents):
            agent = agent_entry[0]
            agent.set_random_seed(p_seed)

    
## -------------------------------------------------------------------------------------------------
    def compute_action(self, p_state: State) -> Action:
        self.log(self.C_LOG_TYPE_I, 'Start of action computation for all agents...')

        action = Action()

        for agent, weight in self._agents:
            action_agent = agent.compute_action(p_state)
            action_element = action_agent.get_elem(agent.get_id())
            action_element.set_weight(weight)
            action.add_elem(agent.get_id(), action_element)

        self.log(self.C_LOG_TYPE_I, 'End of action computation for all agents...')
        return action

    
## -------------------------------------------------------------------------------------------------
    def _adapt(self, *p_args) -> bool:
        state = p_args[0]
        reward = p_args[1]

        self.log(self.C_LOG_TYPE_I, 'Start of adaptation for all agents...')

        adapted = False
        for agent_entry in self._agents:
            agent = agent_entry[0]
            if (reward.get_type() != Reward.C_TYPE_OVERALL) and not reward.is_rewarded(agent.get_id()):
                continue
            self.log(self.C_LOG_TYPE_I, 'Start adaption for agent', agent.get_id())
            adapted = agent.adapt(state, reward) or adapted

        self.log(self.C_LOG_TYPE_I, 'End of adaptation for all agents...')

        self._set_adapted(adapted)
        return adapted

    
## -------------------------------------------------------------------------------------------------
    def clear_buffer(self):
        for agent_entry in self._agents:
            agent_entry[0].clear_buffer()

    
## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        """
        Doesn't support embedded plot of underlying agent hierarchy.
        """

        self.log(self.C_LOG_TYPE_I, 'Init vizualization for all agents...')

        for agent_entry in self._agents:
            agent_entry[0].init_plot(None)

    
## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        self.log(self.C_LOG_TYPE_I, 'Start vizualization for all agents...')

        for agent_entry in self._agents:
            agent_entry[0].update_plot()



## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro.pool.policies
## -- Module  : dummy
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-08  1.0.0     MRD      Creation
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.0 (2021-12-08)

This module provide Policy Dummy for unittest purpose.
"""

from mlpro.rl.models import *
from mlpro.rl.pool.sarsbuffer.RandomSARSBuffer import RandomSARSBuffer

class MyDummyPolicy(Policy):
    """
    Creates a policy that satisfies mlpro interface.
    """
    C_NAME          = 'MyPolicy'
    C_BUFFER_CLS    = RandomSARSBuffer

    def __init__(self, p_observation_space:MSpace, p_action_space:MSpace, p_buffer_size, p_batch_size=5, p_warm_up_step=10, p_ada=True, p_logging=True):
        """
         Parameters:
            p_state_space       State space object
            p_action_space      Action space object
            p_ada               Boolean switch for adaptivity
            p_logging           Boolean switch for logging functionality
        """

        super().__init__(p_observation_space, p_action_space, p_buffer_size=p_buffer_size, p_ada=p_ada, p_logging=p_logging)
        self._state_space   = p_observation_space
        self._action_space  = p_action_space
        self.warm_up_phase = p_warm_up_step
        self.batch_size = p_batch_size
        self.additional_buffer_element = {}
        self.set_id(0)

    def add_buffer(self, p_buffer_element: SARSElement):
        """
        Intended to save the data to the buffer. By default it save the SARS data.
        
        """
        buffer_element = self._add_additional_buffer(p_buffer_element)
        self._buffer.add_element(buffer_element)

    def _add_additional_buffer(self, p_buffer_element: SARSElement):
        p_buffer_element.add_value_element(self.additional_buffer_element)
        return p_buffer_element

    def clear_buffer(self):
        self._buffer.clear()

    def compute_action(self, p_state: State) -> Action:
        # 1 Create a numpy array for your action values 
        my_action_values = np.zeros(self._action_space.get_num_dim())

        # 2 Computing action values is up to you...
        for d in range(self._action_space.get_num_dim()):
            my_action_values[d] = random.random() 

        # 3 Return an action object with your values
        return Action(self._id, self._action_space, my_action_values)


    def _adapt(self, *p_args) -> bool:
        # 1 Adapting the internal policy is up to you...
        # Add data to buffer
        self.add_buffer(p_args[0])

        if len(self._buffer) < self.warm_up_phase:
            self.log(self.C_LOG_TYPE_I, 'Buffer is not full yet, keep collecting data!')
            return False

        sar_data = self._buffer.get_sample(self.batch_size)

        # 2 Only return True if something has been adapted...
        return True


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.dsm
## -- Module  : models
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2022-01-06  0.0.0     DA       Creation
## -------------------------------------------------------------------------------------------------

"""
Ver. 0.0.0 (2022-01-06)

Model classes for efficient online adaptive data stream processing.
"""


from time import CLOCK_THREAD_CPUTIME_ID
from mlpro.bf.various import *
from mlpro.bf.ml import *




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class ProcessingStep(Model):
    """
    Model class for an adaptive data stream (pre-)processing step.
    """

    C_TYPE          = 'Prepro Step'

## -------------------------------------------------------------------------------------------------
    def process_step(self, p_x_add, p_x_del): 
        """
        Processes step.

        Parameters:
            p_x_add     List of input vectors to be added
            p_x_del     List of input vectors to be deleted

        Returns: 
            Nothing
        """

        # 1 Processing steps before policy adaption
        self.log('Process custom step after adaption')
        self.process_before(p_x_add, p_x_del)

        # 2 Policy adaption
#        self.log('Policy adaption')
#        self.policy_adapted = self.adapt_policy(p_x_add, p_x_del)
#        self.log('Policy adapted = ' + self.policy_adapted)

        # 3 Processing steps after policy adaption
        self.log('Process custom step after adaption')
        self.process_after(p_x_add, p_x_del)


## -------------------------------------------------------------------------------------------------
    def process_before(self, p_x_add, p_x_del): 
        """
        Processes custom steps before policy adaption. To be redefined.

        Parameters:
            p_x_add     List of input vectors to be added
            p_x_del     List of input vectors to be deleted

        Returns: 
            Nothing
        """

        pass


## -------------------------------------------------------------------------------------------------
    def process_after(self, p_x_add, p_x_del): 
        """
        Processes custom steps after policy adaption. To be redefined.

        Parameters:
            p_x_add     List of input vectors to be added
            p_x_del     List of input vectors to be deleted

        Returns: 
            Nothing
        """

        pass





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class StreamProcessor(ProcessingStep):
    """
    Model class for sequential adaptive stream processing with optional data preprocessing. Owhn 
    policy adaption and processing steps can be implemented by redefining methods adapt_policy() and
    process_custom().
    """

    C_TYPE          = 'Stream Processor'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_ada=True, p_logging=Log.C_LOG_ALL):
        super().__init__(p_buffer_size=0, p_ada=p_ada, p_logging=p_logging)
        self.prepro_steps   = []
 

## -------------------------------------------------------------------------------------------------
    def switch_logging(self, p_logging):
        Log.switch_logging(self, p_logging=p_logging)
        for step in self.prepro_steps:
            step.switch_logging(p_logging=p_logging)


## -------------------------------------------------------------------------------------------------
    def add_prepro_step(self, p_step:ProcessingStep):
        """
        Adds a preprocessing step.

        Parameters:
            p_step      Preprocessing step object to be added
 
        Returns: 
            Nothing
        """

        p_step.set_adaptivity(self.adaptivity)
        p_step.switch_logging(p_logging=self.logging)
        self.prepro_steps.append(p_step)


## -------------------------------------------------------------------------------------------------
    def process(self, p_x):
        """
        Processes input in three phases: at first all preprocessing steps will be executed. After that
        the own policy will be adapted and at last the own process steps will be executed.

        Parameters:
            p_x         Input vector x

        Returns: 
            Nothing
        """

        # 0 Intro
        x_add   = []
        x_del   = []
        x_add.append(p_x)
        self.log(self.C_LOG_TYPE_I, 'Start processing of input ', p_x)

        # 1 Preprocessing
        if len(self.prepro_steps) > 0:
            self.log(self.C_LOG_TYPE_I, 'Start of preprocessing')

            for step_id, step in enumerate(self.prepro_steps): 
                self.log(self.C_LOG_TYPE_I, 'Preprocessing step ' + str(step_id) + ': ' + step.C_TYPE + ' ' + step.C_NAME)
                step.process_step(x_add, x_del)

            self.log('End of preprocessing')

        
        # 2 Adaption of own policy and main processing
        self.process_step(x_add, x_del)





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Stream:

    C_NAME          = '????'
    C_DESCRIPTION   = '...'
    C_URL           = ''
    C_CITATION      = ''
    C_DOI           = ''
    C_FEATURES      = 0
    C_INSTANCES     = 0

## -------------------------------------------------------------------------------------------------
    def __init__(self) -> None:
        self.def_space()
        self.reset()


## -------------------------------------------------------------------------------------------------
    def def_space(self):
        """
        Defines the internal feature space and it's dimensions. To be redefined. Please bind a well
        defined space object to the internal attribute self.space. 
        """
        
        # Example implementation using the Euclidian space...
        self.space = ESpace()
        self.space.add_dim(Dimension(0, 'X1', '', '', 'm', '', [-20,20]))
        self.space.add_dim(Dimension(1, 'X2', '', '', 'm/s', '', [-10,10]))
        self.space.add_dim(Dimension(2, 'X3', '', '', 'm/s²', '', [-5,5]))


## -------------------------------------------------------------------------------------------------
    def get_space(self):
        return self.space


## -------------------------------------------------------------------------------------------------
    def reset(self):
        """
        Resets stream generator. To be redefined.
        """

        pass


## -------------------------------------------------------------------------------------------------
    def get_next(self):
        """
        Returns next data stream instance or None at the end of the stream. To be redefined.
        """

        return None





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class StreamProcess(Log):
    """
    Stream process, consisting of stream and stream processor object.
    """

    C_TYPE      = 'Stream Process'
    C_NAME      = ''

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_stream:Stream, p_sproc:StreamProcessor, p_logging=True):
        """
        Parameters:
            p_stream        Stream object
            p_sproc         Stream processor object
            p_logging       Boolean switch for logging
        """

        super().__init__(p_logging=p_logging)
        self.stream = p_stream
        self.sproc  = p_sproc

        self.log(self.C_LOG_TYPE_I, 'Stream', self.stream.C_NAME, ' registered')
        self.log(self.C_LOG_TYPE_I, self.sproc.C_TYPE, self.sproc.C_NAME, ' registered')


## -------------------------------------------------------------------------------------------------
    def run(self, p_inst_limit=0, p_feature_ids=None):
        """
        Reads and processes all/limited number of stream instances.

        Parameters:
            p_inst_limit    Optional limitation of instances.
            p_feature_ids   Optional list of ids of features to be processed

        Returns:
            Number of processed instances.
        """

        # 1 Intro
        self.log(self.C_LOG_TYPE_I, 'Start of stream processing (limit='+ str(p_inst_limit) + ')')
        num_inst = 0


        # 2 Main processing loop
        while True:
            inst = self.stream.get_next()
            if inst == None:
                self.log(self.C_LOG_TYPE_I, 'Stream limit reached') 
                break

            num_inst += 1
            self.sproc.process(inst)
            self.log(self.C_LOG_TYPE_I, 'Instance', inst, 'processed')
            if ( p_inst_limit > 0 ) and ( num_inst == p_inst_limit ): break          


        # 3 Outro
        self.log(self.C_LOG_TYPE_I, 'End of stream processing (' + str(num_inst) + ' instances)')
        return num_inst





## -------------------------------------------------------------------------------------------------
## -- Class Group: Special types of preprocessing steps
## -------------------------------------------------------------------------------------------------


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DataWindow (ProcessingStep):
    """
    Model class for data windows that can be used to deal with concept drift. 
    """

    C_TYPE      = 'Data Window'





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Normalization (ProcessingStep):
    """
    Model class for adaptive normalization algorithms. 
    """

    C_TYPE      = 'Normalization'

## -------------------------------------------------------------------------------------------------
    def process_before(self, p_x_add, p_x_del):
        self.backup_policy()


## -------------------------------------------------------------------------------------------------
    def process_after(self, p_x_add, p_x_del):
        for x in p_x_add:
            x = self.normalize(x)


## -------------------------------------------------------------------------------------------------
    def normalize(self, p_x):
        """
        Normalizes an input vector - either by using the recent or the backup policy. To be redefined.

        Parameters:
            p_x         Input vector x to be denormalized
        
        Returns:
            Normalized input vector.
        """
        
        pass   


## -------------------------------------------------------------------------------------------------
    def denormalize(self, p_x, p_backup=True):
        """
        Denormalizes an input - either by using the recent or the backup policy. To be redefined.

        Parameters:
            p_x         Input vector x to be denormalized
            p_backup    If True, the backup policy shall be used. Recent policy otherwise.

        Returns:
            Denormalized input vector.    
        """
        
        pass


## -------------------------------------------------------------------------------------------------
    def renormalize(self, p_x):
        """
        Reormalizes an input vector by denormalizing it with the backup policy and normalizing it
        with the recent policy after that.

        Parameters:
            p_x         Input vector x to be denormalized

        Returns:
            Reormalized input vector.    
        """

        return self.normalize(self.denormalize(p_x, p_backup=True))



## -------------------------------------------------------------------------------------------------
    def backup_policy(self):
        """
        Backups the recent policy. To be redefined.
        """

        pass




## -------------------------------------------------------------------------------------------------
## -- Class Group: Special types of stream processing applications
## -------------------------------------------------------------------------------------------------





## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.sl.afct
## -- Module  : afctrans_pytorch
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-17  0.0.0     MRD       Creation
## -- 2021-12-17  1.0.0     MRD       Released first version
## -- 2022-01-02  2.0.0     MRD       Re-released afct for pytorch
## -- 2022-05-22  2.0.1     MRD       Renamed Class to TorchAFct
## -------------------------------------------------------------------------------------------------

"""
Ver. 2.0.1 (2022-05-22)

This module provides Adaptive Functions for state transition with Neural Network based on Pytorch.
"""

import torch

from mlpro.rl.models import *

class TorchBuffer(Buffer, torch.utils.data.Dataset):
    def __init__(self, p_size=1):
        Buffer.__init__(self, p_size=p_size)
        self._internal_counter = 0

    def add_element(self, p_elem: BufferElement):
        Buffer.add_element(self, p_elem)
        self._internal_counter += 1

    def get_internal_counter(self):
        return self._internal_counter

    def __getitem__(self,idx):
        return self._data_buffer["input"][idx], self._data_buffer["output"][idx]


class TorchBufferElement(BufferElement):
    def __init__(self, p_input: torch.Tensor, p_output: torch.Tensor):

        super().__init__({"input": p_input, "output": p_output})


class TorchAFct(AdaptiveFunction):
    C_NAME = "Pytorch based Adaptive Function"

    def __init__(
        self,
        p_input_space: MSpace,
        p_output_space: MSpace,
        p_output_elem_cls=Element,
        p_threshold=0,
        p_buffer_size=0,
        p_ada=True,
        p_logging=Log.C_LOG_ALL,
        **p_par
    ):

        super().__init__(
            p_input_space=p_input_space,
            p_output_space=p_output_space,
            p_output_elem_cls=p_output_elem_cls,
            p_threshold=p_threshold,
            p_buffer_size=p_buffer_size,
            p_ada=p_ada,
            p_logging=p_logging,
            **p_par
        )

        self._setup_model()

        if self.net_model is None:
            raise ParamError("Please assign your network model to self.net_model")
        
    
    def _setup_model(self):
        """
        Setup Neural Network.
        
        Here, the user needs to implement the Neural Network structure and all the necessary variable
        to train the neural network, e.g, optimizer, loss function.

        Input space can be obtained from the following variable.
        self._input_space

        Output space can be obtained from the following variable.
        self._output_space

        Please return the neural network.
        self.net_model = neural_network_model
        """

        return None

    def input_preproc(self, p_input: Element) -> torch.Tensor:
        # Convert p_input from Element to Tensor
        input = torch.Tensor([p_input.get_values()])

        # Preprocessing Data if needed
        input = self._input_preproc(input)

        return input

    def output_postproc(self, p_output: torch.Tensor) -> list:
        # Output Post Processing
        output = self._output_postproc(p_output)

        # Convert output from Tensor to List
        output = output.detach().flatten().tolist()

        return output

    def _input_preproc(self, p_input: torch.Tensor) -> torch.Tensor:
        return p_input

    def _output_postproc(self, p_output: torch.Tensor) -> torch.Tensor:
        return p_output

    def _map(self, p_input: Element, p_output: Element):
        # Input pre processing
        input = self.input_preproc(p_input)

        # Make prediction
        output = self.net_model(input)

        # Output post processing
        output = self.output_postproc(output)

        p_output.set_values(output)


#!/usr/bin/env python

import rospy
from gazebo_ros_link_attacher.srv import Attach, AttachRequest, AttachResponse


if __name__ == '__main__':
    rospy.init_node('demo_attach_links')
    rospy.loginfo("Creating ServiceProxy to /link_attacher_node/attach")
    attach_srv = rospy.ServiceProxy('/link_attacher_node/attach',
                                    Attach)
    attach_srv.wait_for_service()
    rospy.loginfo("Created ServiceProxy to /link_attacher_node/attach")

    # Link them
    rospy.loginfo("Attaching cube1 and cube2")
    req = AttachRequest()
    req.model_name_1 = "cube1"
    req.link_name_1 = "link"
    req.model_name_2 = "cube2"
    req.link_name_2 = "link"

    attach_srv.call(req)
    # From the shell:
    """
rosservice call /link_attacher_node/attach "model_name_1: 'cube1'
link_name_1: 'link'
model_name_2: 'cube2'
link_name_2: 'link'"
    """

    rospy.loginfo("Attaching cube2 and cube3")
    req = AttachRequest()
    req.model_name_1 = "cube2"
    req.link_name_1 = "link"
    req.model_name_2 = "cube3"
    req.link_name_2 = "link"

    attach_srv.call(req)

